# pylint: disable=no-name-in-module
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Type, TypeVar, cast

import shiboken6 as shb
from bec_lib import bec_logger
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QLabel,
    QLineEdit,
    QSlider,
    QSpinBox,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.widgets.utility.toggle.toggle import ToggleSwitch

if TYPE_CHECKING:  # pragma: no cover
    from bec_widgets.utils import BECConnector

logger = bec_logger.logger

TAncestor = TypeVar("TAncestor", bound=QWidget)


@dataclass(frozen=True)
class WidgetTreeNode:
    widget: QWidget
    parent: QWidget | None
    depth: int
    prefix: str


class WidgetHandler(ABC):
    """Abstract base class for all widget handlers."""

    @abstractmethod
    def get_value(self, widget: QWidget, **kwargs):
        """Retrieve value from the widget instance."""

    @abstractmethod
    def set_value(self, widget: QWidget, value):
        """Set a value on the widget instance."""

    def connect_change_signal(self, widget: QWidget, slot):
        """
        Connect a change signal from this widget to the given slot.
        If the widget type doesn't have a known "value changed" signal, do nothing.

        slot: a function accepting two arguments (widget, value)
        """
        pass


class LineEditHandler(WidgetHandler):
    """Handler for QLineEdit widgets."""

    def get_value(self, widget: QLineEdit, **kwargs) -> str:
        return widget.text()

    def set_value(self, widget: QLineEdit, value: str) -> None:
        widget.setText(value)

    def connect_change_signal(self, widget: QLineEdit, slot):
        widget.textChanged.connect(lambda text, w=widget: slot(w, text))


class ComboBoxHandler(WidgetHandler):
    """Handler for QComboBox widgets."""

    def get_value(self, widget: QComboBox, as_string: bool = False, **kwargs) -> int | str:
        if as_string is True:
            return widget.currentText()
        return widget.currentIndex()

    def set_value(self, widget: QComboBox, value: int | str) -> None:
        if isinstance(value, str):
            value = widget.findText(value)
        if isinstance(value, int):
            widget.setCurrentIndex(value)

    def connect_change_signal(self, widget: QComboBox, slot):
        # currentIndexChanged(int) or currentIndexChanged(str) both possible.
        # We use currentIndexChanged(int) for a consistent behavior.
        widget.currentIndexChanged.connect(lambda idx, w=widget: slot(w, self.get_value(w)))


class TableWidgetHandler(WidgetHandler):
    """Handler for QTableWidget widgets."""

    def get_value(self, widget: QTableWidget, **kwargs) -> list:
        return [
            [
                widget.item(row, col).text() if widget.item(row, col) else ""
                for col in range(widget.columnCount())
            ]
            for row in range(widget.rowCount())
        ]

    def set_value(self, widget: QTableWidget, value) -> None:
        for row, row_values in enumerate(value):
            for col, cell_value in enumerate(row_values):
                item = QTableWidgetItem(str(cell_value))
                widget.setItem(row, col, item)

    def connect_change_signal(self, widget: QTableWidget, slot):
        # If desired, we could connect cellChanged(row, col) and then fetch all data.
        # This might be noisy if table is large.
        # For demonstration, connect cellChanged to update entire table value.
        def on_cell_changed(row, col, w=widget):
            val = self.get_value(w)
            slot(w, val)

        widget.cellChanged.connect(on_cell_changed)


class SpinBoxHandler(WidgetHandler):
    """Handler for QSpinBox and QDoubleSpinBox widgets."""

    def get_value(self, widget: QSpinBox | QDoubleSpinBox, **kwargs):
        return widget.value()

    def set_value(self, widget: QSpinBox | QDoubleSpinBox, value):
        widget.setValue(value)

    def connect_change_signal(self, widget: QSpinBox | QDoubleSpinBox, slot):
        widget.valueChanged.connect(lambda val, w=widget: slot(w, val))


class CheckBoxHandler(WidgetHandler):
    """Handler for QCheckBox widgets."""

    def get_value(self, widget: QCheckBox, **kwargs):
        return widget.isChecked()

    def set_value(self, widget: QCheckBox, value):
        widget.setChecked(value)

    def connect_change_signal(self, widget: QCheckBox, slot):
        widget.toggled.connect(lambda val, w=widget: slot(w, val))


class SlideHandler(WidgetHandler):
    """Handler for QCheckBox widgets."""

    def get_value(self, widget: QSlider, **kwargs):
        return widget.value()

    def set_value(self, widget: QSlider, value):
        widget.setValue(value)

    def connect_change_signal(self, widget: QSlider, slot):
        widget.valueChanged.connect(lambda val, w=widget: slot(w, val))


class ToggleSwitchHandler(WidgetHandler):
    """Handler for ToggleSwitch widgets."""

    def get_value(self, widget: ToggleSwitch, **kwargs):
        return widget.checked

    def set_value(self, widget: ToggleSwitch, value):
        widget.checked = value

    def connect_change_signal(self, widget: ToggleSwitch, slot):
        widget.enabled.connect(lambda val, w=widget: slot(w, val))


class LabelHandler(WidgetHandler):
    """Handler for QLabel widgets."""

    def get_value(self, widget: QLabel, **kwargs):
        return widget.text()

    def set_value(self, widget: QLabel, value):
        widget.setText(value)

    # QLabel typically doesn't have user-editable changes. No signal to connect.
    # If needed, this can remain empty.


class WidgetIO:
    """Public interface for getting, setting values and connecting signals using handler mapping"""

    _handlers = {
        QLineEdit: LineEditHandler,
        QComboBox: ComboBoxHandler,
        QTableWidget: TableWidgetHandler,
        QSpinBox: SpinBoxHandler,
        QDoubleSpinBox: SpinBoxHandler,
        QCheckBox: CheckBoxHandler,
        QLabel: LabelHandler,
        ToggleSwitch: ToggleSwitchHandler,
        QSlider: SlideHandler,
    }

    @staticmethod
    def get_value(widget, ignore_errors=False, **kwargs):
        """
        Retrieve value from the widget instance.

        Args:
            widget: Widget instance.
            ignore_errors(bool, optional): Whether to ignore if no handler is found.
        """
        handler_class = WidgetIO._find_handler(widget)
        if handler_class:
            return handler_class().get_value(widget, **kwargs)  # Instantiate the handler
        if not ignore_errors:
            raise ValueError(f"No handler for widget type: {type(widget)}")
        return None

    @staticmethod
    def set_value(widget, value, ignore_errors=False):
        """
        Set a value on the widget instance.

        Args:
            widget: Widget instance.
            value: Value to set.
            ignore_errors(bool, optional): Whether to ignore if no handler is found.
        """
        handler_class = WidgetIO._find_handler(widget)
        if handler_class:
            handler_class().set_value(widget, value)  # Instantiate the handler
        elif not ignore_errors:
            raise ValueError(f"No handler for widget type: {type(widget)}")

    @staticmethod
    def connect_widget_change_signal(widget, slot):
        """
        Connect the widget's value-changed signal to a generic slot function (widget, value).
        This now delegates the logic to the widget's handler.
        """
        handler_class = WidgetIO._find_handler(widget)
        if handler_class:
            handler = handler_class()
            handler.connect_change_signal(widget, slot)

    @staticmethod
    def check_and_adjust_limits(spin_box: QDoubleSpinBox, number: float):
        """
        Check if the new limits are within the current limits, if not adjust the limits.

        Args:
            number(float): The new value to check against the limits.
        """

        min_value = spin_box.minimum()
        max_value = spin_box.maximum()

        # Calculate the new limits
        new_limit = number + 5 * number

        if number < min_value:
            spin_box.setMinimum(new_limit)
        elif number > max_value:
            spin_box.setMaximum(new_limit)

    @staticmethod
    def _find_handler(widget):
        """
        Find the appropriate handler for the widget by checking its base classes.

        Args:
            widget: Widget instance.

        Returns:
            handler_class: The handler class if found, otherwise None.
        """
        for base in type(widget).__mro__:
            if base in WidgetIO._handlers:
                return WidgetIO._handlers[base]
        return None

    @staticmethod
    def find_widgets(widget_class: QWidget | str, recursive: bool = True) -> list[QWidget]:
        """
        Return widgets matching the given class (or class-name string).

        Args:
            widget_class: Either a QWidget subclass or its class-name as a string.
            recursive: If True (default), traverse all top-level widgets and their children;
                       if False, scan app.allWidgets() for a flat list.

        Returns:
            List of QWidget instances matching the class or class-name.
        """
        app = QApplication.instance()
        if app is None:
            raise RuntimeError("No QApplication instance found.")

        # Match by class-name string
        if isinstance(widget_class, str):
            name = widget_class
            if recursive:
                result: list[QWidget] = []
                for top in app.topLevelWidgets():
                    if top.__class__.__name__ == name:
                        result.append(top)
                    result.extend(
                        w for w in top.findChildren(QWidget) if w.__class__.__name__ == name
                    )
                return result
            return [w for w in app.allWidgets() if w.__class__.__name__ == name]

        # Match by actual class
        if recursive:
            result: list[QWidget] = []
            for top in app.topLevelWidgets():
                if isinstance(top, widget_class):
                    result.append(top)
                result.extend(top.findChildren(widget_class))
            return result

        return [w for w in app.allWidgets() if isinstance(w, widget_class)]


################## for exporting and importing widget hierarchies ##################


class WidgetHierarchy:
    @staticmethod
    def iter_widget_tree(widget: QWidget, *, exclude_internal_widgets: bool = True):
        """
        Yield WidgetTreeNode entries for the widget hierarchy.
        """
        visited: set[int] = set()
        yield from WidgetHierarchy._iter_widget_tree_nodes(
            widget, None, exclude_internal_widgets, visited, [], 0
        )

    @staticmethod
    def _iter_widget_tree_nodes(
        widget: QWidget,
        parent: QWidget | None,
        exclude_internal_widgets: bool,
        visited: set[int],
        branch_flags: list[bool],
        depth: int,
    ):
        if widget is None or not shb.isValid(widget):
            return
        widget_id = id(widget)
        if widget_id in visited:
            return
        visited.add(widget_id)

        prefix = WidgetHierarchy._build_prefix(branch_flags)
        yield WidgetTreeNode(widget=widget, parent=parent, depth=depth, prefix=prefix)

        children = WidgetHierarchy._filtered_children(widget, exclude_internal_widgets)
        for idx, child in enumerate(children):
            is_last = idx == len(children) - 1
            yield from WidgetHierarchy._iter_widget_tree_nodes(
                child,
                widget,
                exclude_internal_widgets,
                visited,
                branch_flags + [is_last],
                depth + 1,
            )

    @staticmethod
    def _build_prefix(branch_flags: list[bool]) -> str:
        if not branch_flags:
            return ""
        parts: list[str] = []
        for flag in branch_flags[:-1]:
            parts.append("   " if flag else "│  ")
        parts.append("└─ " if branch_flags[-1] else "├─ ")
        return "".join(parts)

    @staticmethod
    def _filtered_children(widget: QWidget, exclude_internal_widgets: bool) -> list[QWidget]:
        children: list[QWidget] = []
        for child in widget.findChildren(QWidget, options=Qt.FindDirectChildrenOnly):
            if not shb.isValid(child):
                continue
            if (
                exclude_internal_widgets
                and isinstance(widget, QComboBox)
                and child.__class__.__name__ in ["QFrame", "QBoxLayout", "QListView"]
            ):
                continue
            children.append(child)
        return children

    @staticmethod
    def print_widget_hierarchy(
        widget,
        indent: int = 0,
        grab_values: bool = False,
        prefix: str = "",
        exclude_internal_widgets: bool = True,
        only_bec_widgets: bool = False,
        show_parent: bool = True,
    ) -> None:
        """
        Print the widget hierarchy to the console.

        Args:
            widget: Widget to print the hierarchy of.
            indent(int, optional): Level of indentation.
            grab_values(bool,optional): Whether to grab the values of the widgets.
            prefix(str,optional): Custom string prefix for indentation.
            exclude_internal_widgets(bool,optional): Whether to exclude internal widgets (e.g. QComboBox in PyQt6).
            only_bec_widgets(bool, optional): Whether to print only widgets that are instances of BECWidget.
            show_parent(bool, optional): Whether to display which BECWidget is the parent of each discovered BECWidget.
        """
        from bec_widgets.utils import BECConnector
        from bec_widgets.widgets.plots.waveform.waveform import Waveform

        for node in WidgetHierarchy.iter_widget_tree(
            widget, exclude_internal_widgets=exclude_internal_widgets
        ):
            current = node.widget
            is_bec = isinstance(current, BECConnector)
            if only_bec_widgets and not is_bec:
                continue

            parent_info = ""
            if show_parent and is_bec:
                ancestor = WidgetHierarchy.get_becwidget_ancestor(current)
                if ancestor:
                    parent_label = ancestor.objectName() or ancestor.__class__.__name__
                    parent_info = f" parent={parent_label}"
                else:
                    parent_info = " parent=None"

            widget_info = f"{current.__class__.__name__} ({current.objectName()}){parent_info}"
            print(node.prefix + widget_info)

            if isinstance(current, Waveform):
                for curve in current.curves:
                    curve_prefix = node.prefix + "   "
                    print(
                        f"{curve_prefix}└─ {curve.__class__.__name__} ({curve.objectName()}) "
                        f"parent={current.objectName()}"
                    )

    @staticmethod
    def print_becconnector_hierarchy_from_app():
        """
        Enumerate ALL BECConnector objects in the QApplication.
        Also detect if a widget is a PlotBase, and add any data items
        (PlotDataItem-like) that are also BECConnector objects.

        Build a parent->children graph where each child's 'parent'
        is its closest BECConnector ancestor. Print the entire hierarchy
        from the root(s).

        The result is a single, consolidated tree for your entire
        running GUI, including PlotBase data items that are BECConnector.
        """
        import sys
        from collections import defaultdict

        from qtpy.QtWidgets import QApplication

        from bec_widgets.utils import BECConnector
        from bec_widgets.widgets.plots.plot_base import PlotBase

        # 1) Gather ALL QWidget-based BECConnector objects
        all_qwidgets = QApplication.allWidgets()
        bec_widgets = set(w for w in all_qwidgets if isinstance(w, BECConnector))

        # 2) Also gather any BECConnector-based data items from PlotBase widgets
        for w in all_qwidgets:
            if isinstance(w, PlotBase) and hasattr(w, "plot_item"):
                plot_item = w.plot_item
                if hasattr(plot_item, "listDataItems"):
                    for data_item in plot_item.listDataItems():
                        if isinstance(data_item, BECConnector):
                            bec_widgets.add(data_item)

        # 3) Build a map of (closest BECConnector parent) -> list of children
        parent_map = defaultdict(list)
        for w in bec_widgets:
            parent_bec = WidgetHierarchy.get_becwidget_ancestor(w)
            parent_map[parent_bec].append(w)

        # 4) Define a recursive printer to show each object's children
        def print_tree(parent, prefix=""):
            children = parent_map[parent]
            for i, child in enumerate(children):
                connector_class = child.__class__.__name__
                connector_name = child.objectName() or connector_class

                if parent is None:
                    parent_label = "None"
                else:
                    parent_label = parent.objectName() or parent.__class__.__name__

                line = f"{connector_class} ({connector_name}) parent={parent_label}"
                # Determine tree-branch symbols
                is_last = i == len(children) - 1
                branch_str = "└─ " if is_last else "├─ "
                print(prefix + branch_str + line)

                # Recurse deeper
                next_prefix = prefix + ("   " if is_last else "│  ")
                print_tree(child, prefix=next_prefix)

        # 5) Print top-level items (roots) whose BECConnector parent is None
        roots = parent_map[None]
        for r_i, root in enumerate(roots):
            root_class = root.__class__.__name__
            root_name = root.objectName() or root_class
            line = f"{root_class} ({root_name}) parent=None"
            is_last_root = r_i == len(roots) - 1
            print(line)
            # Recurse into its children
            print_tree(root, prefix="   ")

    @staticmethod
    def get_becwidget_ancestor(widget):
        """
        Traverse up the parent chain to find the nearest BECConnector.

        Args:
            widget: Starting widget to find the ancestor for.

        Returns:
            The nearest ancestor that is a BECConnector, or None if not found.
        """
        from bec_widgets.utils import BECConnector

        # Guard against deleted/invalid Qt wrappers
        if not shb.isValid(widget):
            return None

        # Retrieve first parent
        parent = widget.parent() if hasattr(widget, "parent") else None
        # Walk up, validating each step
        while parent is not None:
            if not shb.isValid(parent):
                return None
            if isinstance(parent, BECConnector):
                return parent
            parent = parent.parent() if hasattr(parent, "parent") else None
        return None

    @staticmethod
    def export_config_to_dict(
        widget: QWidget,
        config: dict = None,
        indent: int = 0,
        grab_values: bool = False,
        print_hierarchy: bool = False,
        save_all: bool = True,
        exclude_internal_widgets: bool = True,
    ) -> dict:
        """
        Export the widget hierarchy to a dictionary.

        Args:
            widget: Widget to print the hierarchy of.
            config(dict,optional): Dictionary to export the hierarchy to.
            indent(int,optional): Level of indentation.
            grab_values(bool,optional): Whether to grab the values of the widgets.
            print_hierarchy(bool,optional): Whether to print the hierarchy to the console.
            save_all(bool,optional): Whether to save all widgets or only those with values.
            exclude_internal_widgets(bool,optional): Whether to exclude internal widgets (e.g. QComboBox in PyQt6).
        Returns:
            config(dict): Dictionary containing the widget hierarchy.
        """
        if config is None:
            config = {}
        widget_info = f"{widget.__class__.__name__} ({widget.objectName()})"

        # if grab_values and type(widget) in WidgetIO._handlers:
        if grab_values:
            value = WidgetIO.get_value(widget, ignore_errors=True)
            if value is not None or save_all:
                if widget_info not in config:
                    config[widget_info] = {}
                if value is not None:
                    config[widget_info]["value"] = value

        if print_hierarchy:
            WidgetHierarchy.print_widget_hierarchy(widget, indent, grab_values)

        for child in widget.children():
            # Skip internal widgets of QComboBox in PyQt6
            if (
                exclude_internal_widgets
                and isinstance(widget, QComboBox)
                and child.__class__.__name__ in ["QFrame", "QBoxLayout", "QListView"]
            ):
                continue
            child_config = WidgetHierarchy.export_config_to_dict(
                child, None, indent + 1, grab_values, print_hierarchy, save_all
            )
            if child_config or save_all:
                if widget_info not in config:
                    config[widget_info] = {}
                config[widget_info].update(child_config)

        return config

    @staticmethod
    def import_config_from_dict(widget, config: dict, set_values: bool = False) -> None:
        """
        Import the widget hierarchy from a dictionary.

        Args:
            widget: Widget to import the hierarchy to.
            config:
            set_values:
        """
        widget_name = f"{widget.__class__.__name__} ({widget.objectName()})"
        widget_config = config.get(widget_name, {})
        for child in widget.children():
            child_name = f"{child.__class__.__name__} ({child.objectName()})"
            child_config = widget_config.get(child_name)
            if child_config is not None:
                value = child_config.get("value")
                if set_values and value is not None:
                    WidgetIO.set_value(child, value)
                WidgetHierarchy.import_config_from_dict(child, widget_config, set_values)

    @staticmethod
    def get_bec_connectors_from_parent(widget) -> list:
        """
        Return all BECConnector instances whose closest BECConnector ancestor is the given widget,
        including the widget itself if it is a BECConnector.
        """
        from bec_widgets.utils import BECConnector

        connectors: list[BECConnector] = []
        if isinstance(widget, BECConnector):
            connectors.append(widget)
        for child in widget.findChildren(BECConnector):
            if WidgetHierarchy.get_becwidget_ancestor(child) is widget:
                connectors.append(child)
        return connectors

    @staticmethod
    def find_ancestor(
        widget: QWidget | BECConnector, ancestor_class: Type[TAncestor] | str
    ) -> TAncestor | None:
        """
        Find the closest ancestor of the specified class (or class-name string).

        Args:
            widget(QWidget): The starting widget.
            ancestor_class(Type[TAncestor] | str): The ancestor class or class-name string to search for.

        Returns:
            TAncestor | None: The closest ancestor of the specified class, or None if not found.
        """
        if widget is None or not shb.isValid(widget):
            return None

        try:
            from bec_widgets.utils import BECConnector  # local import to avoid cycles

            is_bec_target = False
            if isinstance(ancestor_class, str):
                is_bec_target = ancestor_class == "BECConnector"
            elif isinstance(ancestor_class, type):
                is_bec_target = issubclass(ancestor_class, BECConnector)

            if is_bec_target:
                ancestor = WidgetHierarchy.get_becwidget_ancestor(widget)
                return cast(TAncestor, ancestor)
        except Exception as e:
            logger.error(f"Error importing BECConnector: {e}")

        parent = widget.parent() if hasattr(widget, "parent") else None
        while parent is not None:
            if not shb.isValid(parent):
                return None
            try:
                if isinstance(ancestor_class, str):
                    if parent.__class__.__name__ == ancestor_class:
                        return cast(TAncestor, parent)
                else:
                    if isinstance(parent, ancestor_class):
                        return cast(TAncestor, parent)
            except Exception as e:
                logger.error(f"Error checking ancestor class: {e}")
            parent = parent.parent() if hasattr(parent, "parent") else None
        return None


# Example usage
def hierarchy_example():  # pragma: no cover
    app = QApplication([])

    # Create instance of WidgetHierarchy
    widget_hierarchy = WidgetHierarchy()

    # Create a simple widget hierarchy for demonstration purposes
    main_widget = QWidget()
    layout = QVBoxLayout(main_widget)
    line_edit = QLineEdit(main_widget)
    combo_box = QComboBox(main_widget)
    table_widget = QTableWidget(2, 2, main_widget)
    spin_box = QSpinBox(main_widget)
    layout.addWidget(line_edit)
    layout.addWidget(combo_box)
    layout.addWidget(table_widget)
    layout.addWidget(spin_box)

    # Add text items to the combo box
    combo_box.addItems(["Option 1", "Option 2", "Option 3"])

    main_widget.show()

    # Hierarchy of original widget
    print(30 * "#")
    print(f"Widget hierarchy for {main_widget.objectName()}:")
    print(30 * "#")
    config_dict = widget_hierarchy.export_config_to_dict(
        main_widget, grab_values=True, print_hierarchy=True
    )
    print(30 * "#")
    print(f"Config dict: {config_dict}")

    # Hierarchy of new widget and set values
    new_config_dict = {
        "QWidget ()": {
            "QLineEdit ()": {"value": "New Text"},
            "QComboBox ()": {"value": 1},
            "QTableWidget ()": {"value": [["a", "b"], ["c", "d"]]},
            "QSpinBox ()": {"value": 10},
        }
    }
    widget_hierarchy.import_config_from_dict(main_widget, new_config_dict, set_values=True)
    print(30 * "#")
    config_dict_new = widget_hierarchy.export_config_to_dict(
        main_widget, grab_values=True, print_hierarchy=True
    )
    config_dict_new_reduced = widget_hierarchy.export_config_to_dict(
        main_widget, grab_values=True, print_hierarchy=True, save_all=False
    )
    print(30 * "#")
    print(f"Config dict new FULL: {config_dict_new}")
    print(f"Config dict new REDUCED: {config_dict_new_reduced}")

    app.exec()


def widget_io_signal_example():  # pragma: no cover
    app = QApplication([])

    main_widget = QWidget()
    layout = QVBoxLayout(main_widget)
    line_edit = QLineEdit(main_widget)
    combo_box = QComboBox(main_widget)
    spin_box = QSpinBox(main_widget)
    combo_box.addItems(["Option 1", "Option 2", "Option 3"])

    layout.addWidget(line_edit)
    layout.addWidget(combo_box)
    layout.addWidget(spin_box)

    main_widget.show()

    def universal_slot(w, val):
        print(f"Widget {w.objectName() or w} changed, new value: {val}")

    # Connect all supported widgets through their handlers
    WidgetIO.connect_widget_change_signal(line_edit, universal_slot)
    WidgetIO.connect_widget_change_signal(combo_box, universal_slot)
    WidgetIO.connect_widget_change_signal(spin_box, universal_slot)

    app.exec_()


if __name__ == "__main__":  # pragma: no cover
    # Change example function to test different scenarios

    # hierarchy_example()
    widget_io_signal_example()



"""Module for Device Manager View."""

from qtpy.QtCore import QRect
from qtpy.QtWidgets import QWidget

from bec_widgets.applications.views.device_manager_view.device_manager_widget import (
    DeviceManagerWidget,
)
from bec_widgets.applications.views.view import ViewBase, ViewTourSteps
from bec_widgets.utils.error_popups import SafeSlot


class DeviceManagerView(ViewBase):
    """
    A view for users to manage devices within the application.
    """

    def __init__(
        self,
        parent: QWidget | None = None,
        content: QWidget | None = None,
        *,
        view_id: str | None = None,
        title: str | None = None,
        **kwargs,
    ):
        super().__init__(
            parent=parent,
            content=content,
            view_id=view_id,
            title=title,
            rpc_passthrough_children=False,
            **kwargs,
        )
        self.device_manager_widget = DeviceManagerWidget(
            parent=self, rpc_exposed=False, rpc_passthrough_children=False
        )
        self.set_content(self.device_manager_widget)

    @SafeSlot()
    def on_enter(self) -> None:
        """Called after the view becomes current/visible.

        Default implementation does nothing. Override in subclasses.
        """
        self.device_manager_widget.on_enter()

    def register_tour_steps(self, guided_tour, main_app) -> ViewTourSteps | None:
        """Register Device Manager components with the guided tour.

        Args:
            guided_tour: The GuidedTour instance to register with.
            main_app: The main application instance (for accessing set_current).

        Returns:
            ViewTourSteps | None: Model containing view title and step IDs.
        """
        step_ids = []
        dm_widget = self.device_manager_widget

        # The device_manager_widget is not yet initialized, so we will register
        # tour steps for its uninitialized state.

        # Register Load Current Config button
        def get_load_current():
            main_app.set_current("device_manager")
            if dm_widget._initialized is True:
                return (None, None)
            return (dm_widget.button_load_current_config, None)

        step_id = guided_tour.register_widget(
            widget=get_load_current,
            title="Load Current Config",
            text="Load the current device configuration from the BEC server.",
        )
        step_ids.append(step_id)

        # Register Load Config From File button
        def get_load_file():
            main_app.set_current("device_manager")
            if dm_widget._initialized is True:
                return (None, None)
            return (dm_widget.button_load_config_from_file, None)

        step_id = guided_tour.register_widget(
            widget=get_load_file,
            title="Load Config From File",
            text="Load a device configuration from a YAML file on disk.",
        )
        step_ids.append(step_id)

        ## Register steps for the initialized state
        # Register main device table
        def get_device_table():
            main_app.set_current("device_manager")
            if dm_widget._initialized is False:
                return (None, None)
            return (dm_widget.device_manager_display.device_table_view, None)

        step_id = guided_tour.register_widget(
            widget=get_device_table,
            title="Device Table",
            text="This table displays the config that is prepared to be uploaded to the BEC server. It allows users to review and modify device config settings, and also validate them before uploading to the BEC server.",
        )
        step_ids.append(step_id)

        col_text_mapping = {
            0: "Shows if a device configuration is valid. Automatically validated when adding a new device.",
            1: "Shows if a device is connectable. Validated on demand.",
            2: "Device name, unique across all devices within a config.",
            3: "Device class used to initialize the device on the BEC server.",
            4: "Defines how BEC treats readings of the device during scans. The options are 'monitored', 'baseline', 'async', 'continuous' or 'on_demand'.",
            5: "Defines how BEC reacts if a device readback fails. Options are 'raise', 'retry', or 'buffer'.",
            6: "User-defined tags associated with the device.",
            7: "A brief description of the device.",
            8: "Device is enabled when the configuration is loaded.",
            9: "Device is set to read-only.",
            10: "This flag allows to configure if the 'trigger' method of the device is called during scans.",
        }

        # We have at least one device registered
        def get_device_table_row(column: int):
            main_app.set_current("device_manager")
            if dm_widget._initialized is False:
                return (None, None)
            table = dm_widget.device_manager_display.device_table_view.table
            header = table.horizontalHeader()
            x = header.sectionViewportPosition(column)
            table.horizontalScrollBar().setValue(x)
            # Recompute after scrolling
            x = header.sectionViewportPosition(column)
            w = header.sectionSize(column)
            h = header.height()
            rect = QRect(x, 0, w, h)
            top_left = header.viewport().mapTo(main_app, rect.topLeft())

            return (QRect(top_left, rect.size()), col_text_mapping.get(column, ""))

        for col, text in col_text_mapping.items():
            step_id = guided_tour.register_widget(
                widget=lambda col=col: get_device_table_row(col),
                title=f"{dm_widget.device_manager_display.device_table_view.table.horizontalHeaderItem(col).text()}",
                text=text,
            )
            step_ids.append(step_id)

        if not step_ids:
            return None

        return ViewTourSteps(view_title="Device Manager", step_ids=step_ids)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from bec_qthemes import apply_theme
    from qtpy.QtWidgets import QApplication

    from bec_widgets.applications.main_app import BECMainApp

    app = QApplication(sys.argv)
    apply_theme("dark")

    _app = BECMainApp()
    screen = app.primaryScreen()
    screen_geometry = screen.availableGeometry()
    screen_width = screen_geometry.width()
    screen_height = screen_geometry.height()
    # 70% of screen height, keep 16:9 ratio
    height = int(screen_height * 0.9)
    width = int(height * (16 / 9))

    # If width exceeds screen width, scale down
    if width > screen_width * 0.9:
        width = int(screen_width * 0.9)
        height = int(width / (16 / 9))

    _app.resize(width, height)
    device_manager_view = DeviceManagerView()
    _app.add_view(
        icon="display_settings",
        title="Device Manager",
        view_id="device_manager",
        widget=device_manager_view.device_manager_widget,
        mini_text="DM",
    )
    _app.show()
    sys.exit(app.exec_())



import os

from bec_lib.device import Positioner
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QSizePolicy

from bec_widgets.widgets.control.device_control.positioner_box import PositionerBox


class PositionerControlLine(PositionerBox):
    """A widget that controls a single device."""

    ui_file = "positioner_control_line.ui"
    dimensions = (60, 600)  # height, width

    PLUGIN = True
    ICON_NAME = "switch_left"

    def __init__(self, parent=None, device: Positioner | str | None = None, *args, **kwargs):
        """Initialize the DeviceControlLine.

        Args:
            parent: The parent widget.
            device (Positioner): The device to control.
        """
        self.current_path = os.path.dirname(__file__)
        self._indicator_switch_width = 0
        self._horizontal_indicator_width = 0
        self._vertical_indicator_width = 15
        self._indicator_thickness = 10
        self._indicator_is_horizontal = False
        self._line_height = self.dimensions[0]
        super().__init__(parent=parent, device=device, *args, **kwargs)
        self._configure_line_layout()
        self._update_indicator_orientation()

    def _configure_line_layout(self):
        device_box = self.ui.device_box
        indicator = self.ui.position_indicator

        self.main_layout.setAlignment(Qt.AlignmentFlag(0))
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.ui.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        device_box.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)

        self._line_height = max(
            self.dimensions[0],
            self.ui.minimumSizeHint().height(),
            self.ui.sizeHint().height(),
            device_box.minimumSizeHint().height(),
            device_box.sizeHint().height(),
        )
        device_box.setFixedHeight(self._line_height)
        device_box.setMinimumWidth(self.dimensions[1])
        device_box.setMaximumWidth(16777215)
        self.setFixedHeight(self._line_height)
        self.setMinimumWidth(self.dimensions[1])

        self.ui.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.ui.verticalLayout.setSpacing(0)
        self.ui.readback.setMaximumWidth(16777215)
        self.ui.setpoint.setMaximumWidth(16777215)
        self.ui.step_size.setMaximumWidth(16777215)

        indicator_hint = indicator.minimumSizeHint()
        step_hint = self.ui.step_size.sizeHint()
        self._indicator_thickness = max(indicator_hint.height(), 10)
        self._vertical_indicator_width = max(indicator.minimumWidth(), 15)
        self._horizontal_indicator_width = max(90, step_hint.width())
        base_width = max(device_box.minimumSizeHint().width(), self.dimensions[1])
        self._indicator_switch_width = (
            base_width - self._vertical_indicator_width + self._horizontal_indicator_width
        )

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._update_indicator_orientation()

    def _update_indicator_orientation(self):
        if not hasattr(self, "ui"):
            return

        indicator = self.ui.position_indicator
        available_width = self.ui.device_box.width() or self.width() or self.dimensions[1]
        should_use_horizontal = available_width >= self._indicator_switch_width
        if should_use_horizontal == self._indicator_is_horizontal:
            return

        self._indicator_is_horizontal = should_use_horizontal
        indicator.vertical = not should_use_horizontal

        if should_use_horizontal:
            indicator.setMinimumSize(self._horizontal_indicator_width, self._indicator_thickness)
            indicator.setMaximumHeight(self._indicator_thickness)
            indicator.setMaximumWidth(16777215)
            indicator.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        else:
            indicator.setMinimumSize(self._vertical_indicator_width, self._indicator_thickness)
            indicator.setMaximumSize(self._vertical_indicator_width, 16777215)
            indicator.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred)

        indicator.updateGeometry()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    widget = PositionerControlLine(device="samy")

    widget.show()
    sys.exit(app.exec_())



from __future__ import annotations

from enum import Enum

import numpy as np
import pyqtgraph as pg
from bec_lib import bec_logger
from qtpy.QtCore import QPoint, QPointF, Qt, Signal
from qtpy.QtWidgets import QHBoxLayout, QLabel, QMainWindow, QVBoxLayout, QWidget

from bec_widgets.utils import ConnectionConfig, Crosshair, EntryValidator
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.utils.fps_counter import FPSCounter
from bec_widgets.utils.plot_indicator_items import BECArrowItem, BECTickItem
from bec_widgets.utils.round_frame import RoundedFrame
from bec_widgets.utils.side_panel import SidePanel
from bec_widgets.utils.toolbars.performance import PerformanceConnection, performance_bundle
from bec_widgets.utils.toolbars.toolbar import ModularToolBar
from bec_widgets.utils.widget_state_manager import WidgetStateManager
from bec_widgets.widgets.containers.layout_manager.layout_manager import LayoutManagerWidget
from bec_widgets.widgets.plots.setting_menus.axis_settings import AxisSettings
from bec_widgets.widgets.plots.toolbar_components.axis_settings_popup import (
    AxisSettingsPopupConnection,
    axis_popup_bundle,
)
from bec_widgets.widgets.plots.toolbar_components.mouse_interactions import (
    MouseInteractionConnection,
    mouse_interaction_bundle,
)
from bec_widgets.widgets.plots.toolbar_components.plot_export import (
    PlotExportConnection,
    plot_export_bundle,
)
from bec_widgets.widgets.plots.toolbar_components.roi import RoiConnection, roi_bundle

logger = bec_logger.logger


class BECViewBox(pg.ViewBox):
    sigPaint = Signal()

    def paint(self, painter, opt, widget):
        super().paint(painter, opt, widget)
        self.sigPaint.emit()

    def itemBoundsChanged(self, item):
        self._itemBoundsCache.pop(item, None)
        if (self.state["autoRange"][0] is not False) or (self.state["autoRange"][1] is not False):
            # check if the call is coming from a mouse-move event
            if hasattr(item, "skip_auto_range") and item.skip_auto_range:
                return
            self._autoRangeNeedsUpdate = True
            self.update()


class UIMode(Enum):
    NONE = 0
    POPUP = 1
    SIDE = 2


class PlotBase(BECWidget, QWidget):
    PLUGIN = False
    RPC = False
    BASE_USER_ACCESS = [
        "enable_toolbar",
        "enable_toolbar.setter",
        "enable_side_panel",
        "enable_side_panel.setter",
        "enable_fps_monitor",
        "enable_fps_monitor.setter",
        "set",
        "title",
        "title.setter",
        "x_label",
        "x_label.setter",
        "y_label",
        "y_label.setter",
        "x_limits",
        "x_limits.setter",
        "y_limits",
        "y_limits.setter",
        "x_grid",
        "x_grid.setter",
        "y_grid",
        "y_grid.setter",
        "inner_axes",
        "inner_axes.setter",
        "outer_axes",
        "outer_axes.setter",
        "lock_aspect_ratio",
        "lock_aspect_ratio.setter",
        "auto_range",
        "auto_range_x",
        "auto_range_x.setter",
        "auto_range_y",
        "auto_range_y.setter",
        "x_log",
        "x_log.setter",
        "y_log",
        "y_log.setter",
        "legend_label_size",
        "legend_label_size.setter",
        "minimal_crosshair_precision",
        "minimal_crosshair_precision.setter",
        "screenshot",
    ]
    USER_ACCESS = [*BECWidget.USER_ACCESS, *BASE_USER_ACCESS]

    # Custom Signals
    property_changed = Signal(str, object)
    crosshair_position_changed = Signal(tuple)
    crosshair_position_clicked = Signal(tuple)
    crosshair_coordinates_changed = Signal(tuple)
    crosshair_coordinates_clicked = Signal(tuple)

    def __init__(
        self,
        parent: QWidget | None = None,
        config: ConnectionConfig | None = None,
        client=None,
        gui_id: str | None = None,
        popups: bool = True,
        **kwargs,
    ) -> None:
        if config is None:
            config = ConnectionConfig(widget_class=self.__class__.__name__)
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        # For PropertyManager identification
        self.get_bec_shortcuts()

        # Layout Management
        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.layout_manager = LayoutManagerWidget(parent=self)
        self.layout_manager.layout.setContentsMargins(0, 0, 0, 0)
        self.layout_manager.layout.setSpacing(0)

        # Property Manager
        self.state_manager = WidgetStateManager(self)

        # Entry Validator
        self.entry_validator = EntryValidator(self.dev)

        # Base widgets elements
        self._popups = popups
        self._ui_mode = UIMode.POPUP if popups else UIMode.SIDE
        self.axis_settings_dialog = None
        self.plot_widget = pg.GraphicsLayoutWidget(parent=self)
        self.plot_widget.ci.setContentsMargins(0, 0, 0, 0)
        self.plot_item = pg.PlotItem(viewBox=BECViewBox(enableMenu=True))
        self.plot_widget.addItem(self.plot_item)
        self.plot_item.visible_items = lambda: self.visible_items
        self.side_panel = SidePanel(self, orientation="left", panel_max_width=280)

        # PlotItem Addons
        self.plot_item.addLegend()
        self.crosshair = None
        self.fps_monitor = None
        self.fps_label = QLabel(alignment=Qt.AlignmentFlag.AlignRight)
        self._user_x_label = ""
        self._x_label_suffix = ""
        self._x_axis_units = ""
        self._user_y_label = ""
        self._y_label_suffix = ""
        self._y_axis_units = ""
        self._minimal_crosshair_precision = 3

        # Plot Indicator Items
        self.tick_item = BECTickItem(parent=self, plot_item=self.plot_item)
        self.arrow_item = BECArrowItem(parent=self, plot_item=self.plot_item)

        # Visibility States
        self._toolbar_visible = True
        self._enable_fps_monitor = False
        self._outer_axes_visible = self.plot_item.getAxis("top").isVisible()
        self._inner_axes_visible = self.plot_item.getAxis("bottom").isVisible()

        self.toolbar = ModularToolBar(parent=self, orientation="horizontal")
        self._init_toolbar()

        self._init_ui()

        self._connect_to_theme_change()
        self._update_theme(None)

    def apply_theme(self, theme: str):
        self.round_plot_widget.apply_theme(theme)

    def _init_ui(self):
        self.layout.addWidget(self.layout_manager)
        self.round_plot_widget = RoundedFrame(parent=self, content_widget=self.plot_widget)
        self.round_plot_widget.setProperty("variant", "plot_background")
        self.round_plot_widget.setProperty("frameless", True)

        self.layout_manager.add_widget(self.round_plot_widget)
        self.layout_manager.add_widget_relative(self.fps_label, self.round_plot_widget, "top")
        self.fps_label.hide()
        self.layout_manager.add_widget_relative(self.side_panel, self.round_plot_widget, "left")
        self.layout_manager.add_widget_relative(self.toolbar, self.fps_label, "top")

        self.ui_mode = self._ui_mode  # to initiate the first time

        # PlotItem ViewBox Signals
        self.plot_item.vb.sigStateChanged.connect(self.viewbox_state_changed)

    def _init_toolbar(self):
        self.toolbar.add_bundle(performance_bundle(self.toolbar.components))
        self.toolbar.add_bundle(plot_export_bundle(self.toolbar.components))
        self.toolbar.add_bundle(mouse_interaction_bundle(self.toolbar.components))
        self.toolbar.add_bundle(roi_bundle(self.toolbar.components))
        self.toolbar.add_bundle(axis_popup_bundle(self.toolbar.components))

        self.toolbar.connect_bundle(
            "plot_base", PlotExportConnection(self.toolbar.components, self)
        )
        self.toolbar.connect_bundle(
            "plot_base", PerformanceConnection(self.toolbar.components, self)
        )
        self.toolbar.connect_bundle(
            "plot_base", MouseInteractionConnection(self.toolbar.components, self)
        )
        self.toolbar.connect_bundle("plot_base", RoiConnection(self.toolbar.components, self))
        self.toolbar.connect_bundle(
            "plot_base", AxisSettingsPopupConnection(self.toolbar.components, self)
        )

        # hide some options by default
        self.toolbar.toggle_action_visibility("fps_monitor", False)

        # Get default viewbox state
        self.toolbar.show_bundles(
            ["plot_export", "mouse_interaction", "roi", "performance", "axis_popup"]
        )

    def add_side_menus(self):
        """Adds multiple menus to the side panel."""
        # Setting Axis Widget
        try:
            axis_setting = AxisSettings(parent=self, target_widget=self)
            self.side_panel.add_menu(
                action_id="axis",
                icon_name="settings",
                tooltip="Show Axis Settings",
                widget=axis_setting,
                title="Axis Settings",
            )
        except ValueError:
            return

    def reset_legend(self):
        """In the case that the legend is not visible, reset it to be visible to top left corner"""
        self.plot_item.legend.autoAnchor(50)

    ################################################################################
    # Toggle UI Elements
    ################################################################################
    @property
    def ui_mode(self) -> UIMode:
        """
        Get the UI mode.
        """
        return self._ui_mode

    @ui_mode.setter
    def ui_mode(self, mode: UIMode):
        """
        Set the UI mode.

        Args:
            mode(UIMode): The UI mode to set.
        """
        if not isinstance(mode, UIMode):
            raise ValueError("ui_mode must be an instance of UIMode")
        self._ui_mode = mode

        # Now, apply the new mode:
        if mode == UIMode.POPUP:
            shown_bundles = self.toolbar.shown_bundles
            if "axis_popup" not in shown_bundles:
                shown_bundles.append("axis_popup")
            self.toolbar.show_bundles(shown_bundles)
            self.side_panel.hide()

        elif mode == UIMode.SIDE:
            shown_bundles = self.toolbar.shown_bundles
            if "axis_popup" in shown_bundles:
                shown_bundles.remove("axis_popup")
            self.toolbar.show_bundles(shown_bundles)
            pb_connection = self.toolbar.bundles["axis_popup"].get_connection("plot_base")
            if pb_connection.axis_settings_dialog is not None:
                pb_connection.axis_settings_dialog.close()
                pb_connection.axis_settings_dialog = None
            self.add_side_menus()
            self.side_panel.show()

    @SafeProperty(bool, doc="Enable popups setting dialogs for the plot widget.")
    def enable_popups(self):
        """
        Enable popups setting dialogs for the plot widget.
        """
        return self.ui_mode == UIMode.POPUP

    @enable_popups.setter
    def enable_popups(self, value: bool):
        """
        Set the popups setting dialogs for the plot widget.

        Args:
            value(bool): The value to set.
        """
        if value:
            self.ui_mode = UIMode.POPUP
        else:
            if self.ui_mode == UIMode.POPUP:
                self.ui_mode = UIMode.NONE

    @SafeProperty(bool, doc="Show Side Panel")
    def enable_side_panel(self) -> bool:
        """
        Show Side Panel
        """
        return self.ui_mode == UIMode.SIDE

    @enable_side_panel.setter
    def enable_side_panel(self, value: bool):
        """
        Show Side Panel

        Args:
            value(bool): The value to set.
        """
        if value:
            self.ui_mode = UIMode.SIDE
        else:
            if self.ui_mode == UIMode.SIDE:
                self.ui_mode = UIMode.NONE

    @SafeProperty(bool, doc="Show Toolbar")
    def enable_toolbar(self) -> bool:
        """
        Show Toolbar.
        """
        return self._toolbar_visible

    @enable_toolbar.setter
    def enable_toolbar(self, value: bool):
        """
        Show Toolbar.

        Args:
            value(bool): The value to set.
        """
        self._toolbar_visible = value
        self.toolbar.setVisible(value)

    @SafeProperty(bool, doc="Enable the FPS monitor.")
    def enable_fps_monitor(self) -> bool:
        """
        Enable the FPS monitor.
        """
        return self._enable_fps_monitor

    @enable_fps_monitor.setter
    def enable_fps_monitor(self, value: bool):
        """
        Enable the FPS monitor.

        Args:
            value(bool): The value to set.
        """
        if value == self._enable_fps_monitor:
            return
        if value:
            self.hook_fps_monitor()
        else:
            self.unhook_fps_monitor()

    ################################################################################
    # ViewBox State Signals
    ################################################################################

    def viewbox_state_changed(self):
        """
        Emit a signal when the state of the viewbox has changed.
        Merges the default pyqtgraphs signal states and also CTRL menu toggles.
        """

        viewbox_state = self.plot_item.vb.getState()
        # Range Limits
        x_min, x_max = viewbox_state["targetRange"][0]
        y_min, y_max = viewbox_state["targetRange"][1]
        self.property_changed.emit("x_min", x_min)
        self.property_changed.emit("x_max", x_max)
        self.property_changed.emit("y_min", y_min)
        self.property_changed.emit("y_max", y_max)

        # Grid Toggles

    ################################################################################
    # Plot Properties
    ################################################################################

    def set(self, **kwargs):
        """
        Set the properties of the plot widget.

        Args:
            **kwargs: Keyword arguments for the properties to be set.

        Possible properties:
            - title: str
            - x_label: str
            - y_label: str
            - x_scale: Literal["linear", "log"]
            - y_scale: Literal["linear", "log"]
            - x_lim: tuple
            - y_lim: tuple
            - legend_label_size: int

        """
        property_map = {
            "title": self.title,
            "x_label": self.x_label,
            "y_label": self.y_label,
            "x_limits": self.x_limits,
            "y_limits": self.y_limits,
            "x_grid": self.x_grid,
            "y_grid": self.y_grid,
            "inner_axes": self.inner_axes,
            "outer_axes": self.outer_axes,
            "lock_aspect_ratio": self.lock_aspect_ratio,
            "auto_range_x": self.auto_range_x,
            "auto_range_y": self.auto_range_y,
            "x_log": self.x_log,
            "y_log": self.y_log,
            "legend_label_size": self.legend_label_size,
        }

        for key, value in kwargs.items():
            if key in property_map:
                setattr(self, key, value)
            else:
                logger.warning(f"Property {key} not found.")

    @SafeProperty(str, auto_emit=True, doc="The title of the axes.")
    def title(self) -> str:
        """
        Set title of the plot.
        """
        return self.plot_item.titleLabel.text

    @title.setter
    def title(self, value: str):
        """
        Set title of the plot.

        Args:
            value(str): The title to set.
        """
        self.plot_item.setTitle(value)

    @SafeProperty(str, auto_emit=True, doc="The text of the x label")
    def x_label(self) -> str:
        """
        The set label for the x-axis.
        """
        return self._user_x_label

    @x_label.setter
    def x_label(self, value: str):
        """
        The set label for the x-axis.

        Args:
            value(str): The label to set.
        """
        self._user_x_label = value
        self._apply_x_label()

    @property
    def x_label_suffix(self) -> str:
        """
        A read-only (or internal) suffix automatically appended to the user label.
        Not settable by the user directly from the UI.
        """
        return self._x_label_suffix

    def set_x_label_suffix(self, suffix: str):
        """
        Public or protected method to update the suffix.
        The user code or subclass (Waveform) can call this
        when x_mode changes, but the AxisSettings won't show it.
        """
        self._x_label_suffix = suffix
        self._apply_x_label()

    @property
    def x_label_units(self) -> str:
        """
        The units of the x-axis.
        """
        return self._x_axis_units

    @x_label_units.setter
    def x_label_units(self, units: str):
        """
        The units of the x-axis.

        Args:
            units(str): The units to set.
        """
        self._x_axis_units = units
        self._apply_x_label()

    @property
    def x_label_combined(self) -> str:
        """
        The final label shown on the axis = user portion + suffix + [units].
        """
        units = f" [{self._x_axis_units}]" if self._x_axis_units else ""
        return self._user_x_label + self._x_label_suffix + units

    def _apply_x_label(self):
        """
        Actually updates the pyqtgraph axis label text to
        the combined label. Called whenever user label or suffix changes.
        """
        final_label = self.x_label_combined
        if self.plot_item.getAxis("bottom").isVisible():
            self.plot_item.setLabel("bottom", text=final_label)

    @SafeProperty(str, auto_emit=True, doc="The text of the y label")
    def y_label(self) -> str:
        """
        The set label for the y-axis.
        """
        return self._user_y_label

    @y_label.setter
    def y_label(self, value: str):
        """
        The set label for the y-axis.
        Args:
            value(str): The label to set.
        """
        self._user_y_label = value
        self._apply_y_label()

    @property
    def y_label_suffix(self) -> str:
        """
        A read-only suffix automatically appended to the y label.
        """
        return self._y_label_suffix

    def set_y_label_suffix(self, suffix: str):
        """
        Public method to update the y label suffix.
        """
        self._y_label_suffix = suffix
        self._apply_y_label()

    @property
    def y_label_units(self) -> str:
        """
        The units of the y-axis.
        """
        return self._y_axis_units

    @y_label_units.setter
    def y_label_units(self, units: str):
        """
        The units of the y-axis.

        Args:
            units(str): The units to set.
        """
        self._y_axis_units = units
        self._apply_y_label()

    @property
    def y_label_combined(self) -> str:
        """
        The final y label shown on the axis = user portion + suffix + [units].
        """
        units = f" [{self._y_axis_units}]" if self._y_axis_units else ""
        return self._user_y_label + self._y_label_suffix + units

    def _apply_y_label(self):
        """
        Actually updates the pyqtgraph y axis label text to
        the combined y label. Called whenever y label or suffix changes.
        """
        final_label = self.y_label_combined
        if self.plot_item.getAxis("bottom").isVisible():
            self.plot_item.setLabel("left", text=final_label)

    def _tuple_to_qpointf(self, tuple: tuple | list):
        """
        Helper function to convert a tuple to a QPointF.

        Args:
            tuple(tuple|list): Tuple or list of two numbers.

        Returns:
            QPointF: The tuple converted to a QPointF.
        """
        if len(tuple) != 2:
            raise ValueError("Limits must be a tuple or list of two numbers.")
        min_val, max_val = tuple
        if not isinstance(min_val, (int, float)) or not isinstance(max_val, (int, float)):
            raise TypeError("Limits must be numbers.")
        if min_val > max_val:
            raise ValueError("Minimum limit cannot be greater than maximum limit.")
        return QPoint(*tuple)

    ################################################################################
    # X limits, has to be SaveProperty("QPointF") because of the tuple conversion for designer,
    # the python properties are used for CLI and API for context dialog settings.

    @SafeProperty("QPointF")
    def x_limits(self) -> QPointF:
        """
        Get the x limits of the plot.
        """
        current_lim = self.plot_item.vb.viewRange()[0]
        return QPointF(current_lim[0], current_lim[1])

    @x_limits.setter
    def x_limits(self, value):
        """
        Set the x limits of the plot.

        Args:
            value(QPointF|tuple|list): The x limits to set.
        """
        if isinstance(value, (tuple, list)):
            value = self._tuple_to_qpointf(value)
        self.plot_item.vb.setXRange(value.x(), value.y(), padding=0)

    @property
    def x_lim(self) -> tuple:
        """
        Get the x limits of the plot.
        """
        return (self.x_limits.x(), self.x_limits.y())

    @x_lim.setter
    def x_lim(self, value):
        """
        Set the x limits of the plot.

        Args:
            value(tuple): The x limits to set.
        """
        self.x_limits = value

    @property
    def x_min(self) -> float:
        """
        Get the minimum x limit of the plot.

        """
        return self.x_limits.x()

    @x_min.setter
    def x_min(self, value: float):
        """
        Set the minimum x limit of the plot.

        Args:
            value(float): The minimum x limit to set.
        """
        self.x_limits = (value, self.x_lim[1])

    @property
    def x_max(self) -> float:
        """
        Get the maximum x limit of the plot.
        """
        return self.x_limits.y()

    @x_max.setter
    def x_max(self, value: float):
        """
        Set the maximum x limit of the plot.

        Args:
            value(float): The maximum x limit to set.
        """
        self.x_limits = (self.x_lim[0], value)

    ################################################################################
    # Y limits, has to be SaveProperty("QPointF") because of the tuple conversion for designer,
    # the python properties are used for CLI and API for context dialog settings.

    @SafeProperty("QPointF")
    def y_limits(self) -> QPointF:
        """
        Get the y limits of the plot.
        """
        current_lim = self.plot_item.vb.viewRange()[1]
        return QPointF(current_lim[0], current_lim[1])

    @y_limits.setter
    def y_limits(self, value):
        """
        Set the y limits of the plot.

        Args:
            value(QPointF|tuple|list): The y limits to set.
        """
        if isinstance(value, (tuple, list)):
            value = self._tuple_to_qpointf(value)
        self.plot_item.vb.setYRange(value.x(), value.y(), padding=0)

    @property
    def y_lim(self) -> tuple:
        """
        Get the y limits of the plot.
        """
        return (self.y_limits.x(), self.y_limits.y())

    @y_lim.setter
    def y_lim(self, value):
        """
        Set the y limits of the plot.

        Args:
            value(tuple): The y limits to set.
        """
        self.y_limits = value

    @property
    def y_min(self) -> float:
        """
        Get the minimum y limit of the plot.
        """
        return self.y_limits.x()

    @y_min.setter
    def y_min(self, value: float):
        """
        Set the minimum y limit of the plot.

        Args:
            value(float): The minimum y limit to set.
        """
        self.y_limits = (value, self.y_lim[1])

    @property
    def y_max(self) -> float:
        """
        Get the maximum y limit of the plot.
        """
        return self.y_limits.y()

    @y_max.setter
    def y_max(self, value: float):
        """
        Set the maximum y limit of the plot.

        Args:
            value(float): The maximum y limit to set.
        """
        self.y_limits = (self.y_lim[0], value)

    @SafeProperty(bool, auto_emit=True, doc="Show grid on the x-axis.")
    def x_grid(self) -> bool:
        """
        Show grid on the x-axis.
        """
        return self.plot_item.ctrl.xGridCheck.isChecked()

    @x_grid.setter
    def x_grid(self, value: bool):
        """
        Show grid on the x-axis.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.showGrid(x=value)

    @SafeProperty(bool, auto_emit=True, doc="Show grid on the y-axis.")
    def y_grid(self) -> bool:
        """
        Show grid on the y-axis.
        """
        return self.plot_item.ctrl.yGridCheck.isChecked()

    @y_grid.setter
    def y_grid(self, value: bool):
        """
        Show grid on the y-axis.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.showGrid(y=value)

    @SafeProperty(bool, auto_emit=True, doc="Set X-axis to log scale if True, linear if False.")
    def x_log(self) -> bool:
        """
        Set X-axis to log scale if True, linear if False.
        """
        return bool(self.plot_item.vb.state.get("logMode", [False, False])[0])

    @x_log.setter
    def x_log(self, value: bool):
        """
        Set X-axis to log scale if True, linear if False.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.setLogMode(x=value)

    @SafeProperty(bool, auto_emit=True, doc="Set Y-axis to log scale if True, linear if False.")
    def y_log(self) -> bool:
        """
        Set Y-axis to log scale if True, linear if False.
        """
        return bool(self.plot_item.vb.state.get("logMode", [False, False])[1])

    @y_log.setter
    def y_log(self, value: bool):
        """
        Set Y-axis to log scale if True, linear if False.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.setLogMode(y=value)

    @SafeProperty(bool, auto_emit=True, doc="Show the outer axes of the plot widget.")
    def outer_axes(self) -> bool:
        """
        Show the outer axes of the plot widget.
        """
        return self._outer_axes_visible

    @outer_axes.setter
    def outer_axes(self, value: bool):
        """
        Show the outer axes of the plot widget.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.showAxis("top", value)
        self.plot_item.showAxis("right", value)

        self._outer_axes_visible = value

    @SafeProperty(bool, auto_emit=True, doc="Show inner axes of the plot widget.")
    def inner_axes(self) -> bool:
        """
        Show inner axes of the plot widget.
        """
        return self._inner_axes_visible

    @inner_axes.setter
    def inner_axes(self, value: bool):
        """
        Show inner axes of the plot widget.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.showAxis("bottom", value)
        self.plot_item.showAxis("left", value)

        self._inner_axes_visible = value
        self._apply_x_label()
        self._apply_y_label()

    @SafeProperty(bool, doc="Invert X axis.")
    def invert_x(self) -> bool:
        """
        Invert X axis.
        """
        return self.plot_item.vb.state.get("xInverted", False)

    @invert_x.setter
    def invert_x(self, value: bool):
        """
        Invert X axis.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.vb.invertX(value)

    @SafeProperty(bool, doc="Invert Y axis.")
    def invert_y(self) -> bool:
        """
        Invert Y axis.
        """
        return self.plot_item.vb.state.get("yInverted", False)

    @invert_y.setter
    def invert_y(self, value: bool):
        """
        Invert Y axis.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.vb.invertY(value)

    @SafeProperty(bool, doc="Lock aspect ratio of the plot widget.")
    def lock_aspect_ratio(self) -> bool:
        """
        Lock aspect ratio of the plot widget.
        """
        return bool(self.plot_item.vb.getState()["aspectLocked"])

    @lock_aspect_ratio.setter
    def lock_aspect_ratio(self, value: bool):
        """
        Lock aspect ratio of the plot widget.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.setAspectLocked(value)

    @SafeProperty(bool, doc="Set auto range for the x-axis.")
    def auto_range_x(self) -> bool:
        """
        Set auto range for the x-axis.
        """
        return bool(self.plot_item.vb.getState()["autoRange"][0])

    @auto_range_x.setter
    def auto_range_x(self, value: bool):
        """
        Set auto range for the x-axis.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.enableAutoRange(x=value)

    @SafeProperty(bool, doc="Set auto range for the y-axis.")
    def auto_range_y(self) -> bool:
        """
        Set auto range for the y-axis.
        """
        return bool(self.plot_item.vb.getState()["autoRange"][1])

    @auto_range_y.setter
    def auto_range_y(self, value: bool):
        """
        Set auto range for the y-axis.

        Args:
            value(bool): The value to set.
        """
        self.plot_item.enableAutoRange(y=value)

    def auto_range(self, value: bool = True):
        """
        On demand apply autorange to the plot item based on the visible curves.

        Args:
            value(bool): If True, apply autorange to the visible curves.
        """
        if not value:
            self.plot_item.enableAutoRange(x=False, y=False)
            return
        self._apply_autorange_only_visible_curves()

    @property
    def visible_items(self):
        crosshair_items = []
        if self.crosshair:
            crosshair_items = [
                self.crosshair.v_line,
                self.crosshair.h_line,
                self.crosshair.coord_label,
            ]
        return [
            item
            for item in self.plot_item.items
            if item.isVisible() and item not in crosshair_items
        ]

    def _apply_autorange_only_visible_curves(self):
        """
        Apply autorange to the plot item based on the provided curves.

        Args:
            curves (list): List of curves to apply autorange to.
        """
        visible_items = self.visible_items

        self.plot_item.autoRange(items=visible_items if visible_items else None)

    @SafeProperty(int, doc="The font size of the legend font.")
    def legend_label_size(self) -> int:
        """
        The font size of the legend font.
        """
        if not self.plot_item.legend:
            return
        scale = self.plot_item.legend.scale() * 9
        return scale

    @legend_label_size.setter
    def legend_label_size(self, value: int):
        """
        The font size of the legend font.

        Args:
            value(int): The font size to set.
        """
        if not self.plot_item.legend:
            return
        scale = (
            value / 9
        )  # 9 is the default font size of the legend, so we always scale it against 9
        self.plot_item.legend.setScale(scale)

    ################################################################################
    # FPS Counter
    ################################################################################

    def update_fps_label(self, fps: float) -> None:
        """
        Update the FPS label.

        Args:
            fps(float): The frames per second.
        """
        if self.fps_label:
            self.fps_label.setText(f"FPS: {fps:.2f}")

    def hook_fps_monitor(self):
        """Hook the FPS monitor to the plot."""
        if self.fps_monitor is None:
            self.fps_monitor = FPSCounter(self.plot_item.vb)
            self.fps_label.show()

            self.fps_monitor.sigFpsUpdate.connect(self.update_fps_label)
            self.update_fps_label(0)
        self._enable_fps_monitor = True

    def unhook_fps_monitor(self, delete_label=True):
        """Unhook the FPS monitor from the plot."""
        if self.fps_monitor is not None and delete_label:
            # Remove Monitor
            self.fps_monitor.cleanup()
            self.fps_monitor.deleteLater()
            self.fps_monitor = None
        if self.fps_label is not None:
            # Hide Label
            self.fps_label.hide()
        self._enable_fps_monitor = False

    ################################################################################
    # Crosshair
    ################################################################################

    def hook_crosshair(self) -> None:
        """Hook the crosshair to all plots."""
        if self.crosshair is None:
            self.crosshair = Crosshair(
                self.plot_item, min_precision=self._minimal_crosshair_precision
            )
            self.crosshair.crosshairChanged.connect(self.crosshair_position_changed)
            self.crosshair.crosshairClicked.connect(self.crosshair_position_clicked)
            self.crosshair.coordinatesChanged1D.connect(self.crosshair_coordinates_changed)
            self.crosshair.coordinatesClicked1D.connect(self.crosshair_coordinates_clicked)
            self.crosshair.coordinatesChanged2D.connect(self.crosshair_coordinates_changed)
            self.crosshair.coordinatesClicked2D.connect(self.crosshair_coordinates_clicked)

    def unhook_crosshair(self) -> None:
        """Unhook the crosshair from all plots."""
        if self.crosshair is not None:
            self.crosshair.crosshairChanged.disconnect(self.crosshair_position_changed)
            self.crosshair.crosshairClicked.disconnect(self.crosshair_position_clicked)
            self.crosshair.coordinatesChanged1D.disconnect(self.crosshair_coordinates_changed)
            self.crosshair.coordinatesClicked1D.disconnect(self.crosshair_coordinates_clicked)
            self.crosshair.coordinatesChanged2D.disconnect(self.crosshair_coordinates_changed)
            self.crosshair.coordinatesClicked2D.disconnect(self.crosshair_coordinates_clicked)
            self.crosshair.cleanup()
            self.crosshair.deleteLater()
            self.crosshair = None

    def toggle_crosshair(self) -> None:
        """Toggle the crosshair on all plots."""
        if self.crosshair is None:
            return self.hook_crosshair()

        self.unhook_crosshair()

    @SafeProperty(
        int,
        auto_emit=True,
        doc="Minimum decimal places for crosshair when dynamic precision is enabled.",
    )
    def minimal_crosshair_precision(self) -> int:
        """
        Minimum decimal places for crosshair when dynamic precision is enabled.
        """
        return self._minimal_crosshair_precision

    @minimal_crosshair_precision.setter
    def minimal_crosshair_precision(self, value: int):
        """
        Set the minimum decimal places for crosshair when dynamic precision is enabled.

        Args:
            value(int): The minimum decimal places to set.
        """
        value_int = max(0, int(value))
        self._minimal_crosshair_precision = value_int
        if self.crosshair is not None:
            self.crosshair.min_precision = value_int

    @SafeSlot()
    def reset(self) -> None:
        """Reset the plot widget."""
        if self.crosshair is not None:
            self.crosshair.clear_markers()
            self.crosshair.update_markers()

    def cleanup(self):
        self.toolbar.cleanup()
        self.unhook_crosshair()
        self.unhook_fps_monitor(delete_label=True)
        self.tick_item.cleanup()
        self.arrow_item.cleanup()
        if self.axis_settings_dialog is not None:
            self.axis_settings_dialog.close()
            self.axis_settings_dialog = None
        self.cleanup_pyqtgraph()
        self.round_plot_widget.close()
        super().cleanup()

    def cleanup_pyqtgraph(self, item: pg.PlotItem | None = None):
        """Cleanup pyqtgraph items."""
        if item is None:
            item = self.plot_item
        item.vb.menu.close()
        item.vb.menu.deleteLater()
        item.ctrlMenu.close()
        item.ctrlMenu.deleteLater()


class DemoPlotBase(QMainWindow):  # pragma: no cover:
    def __init__(self):
        super().__init__()
        self.main_widget = QWidget()
        self.setCentralWidget(self.main_widget)
        self.main_widget.layout = QHBoxLayout(self.main_widget)

        self.plot_popup = PlotBase(popups=True)
        self.plot_popup.title = "PlotBase with popups"
        self.plot_side_panels = PlotBase(popups=False)
        self.plot_side_panels.title = "PlotBase with side panels"

        self.plot_popup.plot_item.plot(np.random.rand(100), pen=(255, 0, 0))
        self.plot_side_panels.plot_item.plot(np.random.rand(100), pen=(0, 255, 0))

        self.main_widget.layout.addWidget(self.plot_side_panels)
        self.main_widget.layout.addWidget(self.plot_popup)

        self.resize(1400, 600)


if __name__ == "__main__":  # pragma: no cover:
    import sys

    from qtpy.QtWidgets import QApplication

    from bec_widgets.widgets.containers.main_window.main_window import BECMainWindow

    app = QApplication(sys.argv)
    launch_window = BECMainWindow()
    pb = PlotBase(popups=False)
    launch_window.setCentralWidget(pb)
    launch_window.show()

    sys.exit(app.exec_())



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from bec_qthemes._icon.material_icons import material_icon
from qtpy.QtCore import QSize
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils import UILoader
from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.settings_dialog import SettingWidget
from bec_widgets.widgets.progress.ring_progress_bar.ring import Ring
from bec_widgets.widgets.utility.visual.colormap_widget.colormap_widget import BECColorMapWidget

if TYPE_CHECKING:  # pragma: no cover
    from bec_widgets.widgets.progress.ring_progress_bar.ring_progress_bar import (
        RingProgressBar,
        RingProgressContainerWidget,
    )


class RingCardWidget(QFrame):
    def __init__(self, ring: Ring, container: RingProgressContainerWidget, parent=None):
        super().__init__(parent)

        self.ring = ring
        self.container = container
        self.details_visible = False
        self.setProperty("skip_settings", True)
        self.setFrameShape(QFrame.Shape.StyledPanel)
        self.setObjectName("RingCardWidget")

        bg = self._get_theme_color("BORDER")
        self.setStyleSheet(f"""
            #RingCardWidget {{
                border: 1px solid {bg.name() if bg else '#CCCCCC'};
                border-radius: 4px;
            }}
            """)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(8, 8, 8, 8)
        layout.setSpacing(6)

        self._init_header(layout)
        self._init_details(layout)

        self._init_values()
        self._connect_signals()
        self.mode_combo.setCurrentText(self._get_display_mode_string(self.ring.config.mode))
        self._set_widget_mode_enabled(self.ring.config.mode)

    @staticmethod
    def _get_theme_color(color_name: str) -> QColor | None:
        app = QApplication.instance()
        if not app:
            return
        if not app.theme:
            return
        return app.theme.color(color_name)

    def _init_header(self, parent_layout: QVBoxLayout):
        """Create the collapsible header with basic controls"""
        header = QWidget()
        layout = QHBoxLayout(header)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(8)

        self.expand_btn = QPushButton("▶")
        self.expand_btn.setFixedWidth(24)
        self.expand_btn.clicked.connect(self.toggle_details)

        self.mode_combo = QComboBox()
        self.mode_combo.addItems(["Manual", "Scan Progress", "Device Readback"])
        self.mode_combo.currentTextChanged.connect(self._update_mode)

        delete_btn = QPushButton(material_icon("delete"), "")

        color = self._get_theme_color("ACCENT_HIGHLIGHT")
        delete_btn.setStyleSheet(f"background-color: {color.name() if color else '#CC181E'}")
        delete_btn.clicked.connect(self._delete_self)

        layout.addWidget(self.expand_btn)
        layout.addWidget(QLabel("Mode"))
        layout.addWidget(self.mode_combo)
        layout.addStretch()
        layout.addWidget(delete_btn)

        parent_layout.addWidget(header)

    def _init_details(self, parent_layout: QVBoxLayout):
        """Create the collapsible details area with the UI file"""
        self.details = QWidget()
        self.details.setVisible(False)

        details_layout = QVBoxLayout(self.details)
        details_layout.setContentsMargins(0, 0, 0, 0)

        # Load UI file into details area
        current_path = os.path.dirname(__file__)
        self.ui = UILoader().load_ui(os.path.join(current_path, "ring_settings.ui"), self.details)
        details_layout.addWidget(self.ui)

        parent_layout.addWidget(self.details)

    def toggle_details(self):
        """Toggle visibility of the details area"""
        self.details_visible = not self.details_visible
        self.details.setVisible(self.details_visible)
        self.expand_btn.setText("▼" if self.details_visible else "▶")

    # --------------------------------------------------------

    def _connect_signals(self):
        """Connect UI signals to ring methods"""
        # Data connections
        self.ui.value_spin_box.valueChanged.connect(self.ring.set_value)
        self.ui.min_spin_box.valueChanged.connect(self._update_min_max)
        self.ui.max_spin_box.valueChanged.connect(self._update_min_max)

        # Config connections
        self.ui.start_angle_spin_box.valueChanged.connect(self.ring.set_start_angle)
        self.ui.direction_combo_box.currentIndexChanged.connect(self._update_direction)
        self.ui.line_width_spin_box.valueChanged.connect(self.ring.set_line_width)
        self.ui.background_color_button.color_changed.connect(self.ring.set_background)
        self.ui.ring_color_button.color_changed.connect(self._on_ring_color_changed)
        self.ui.device_combo_box.device_selected.connect(self._on_device_changed)
        self.ui.signal_combo_box.device_signal_changed.connect(self._on_signal_changed)

    def _init_values(self):
        """Initialize UI values from ring config"""
        # Data values
        self.ui.value_spin_box.setRange(-1e6, 1e6)
        self.ui.value_spin_box.setValue(self.ring.config.value)

        self.ui.min_spin_box.setRange(-1e6, 1e6)
        self.ui.min_spin_box.setValue(self.ring.config.min_value)

        self.ui.max_spin_box.setRange(-1e6, 1e6)
        self.ui.max_spin_box.setValue(self.ring.config.max_value)
        self._update_min_max()

        self.ui.device_combo_box.setEditable(True)
        self.ui.signal_combo_box.setEditable(True)

        device, signal = self.ring.config.device, self.ring.config.signal
        if device:
            self.ui.device_combo_box.set_device(device)
        if signal:
            for i in range(self.ui.signal_combo_box.count()):
                data_item = self.ui.signal_combo_box.itemData(i)
                if data_item and data_item.get("obj_name") == signal:
                    self.ui.signal_combo_box.setCurrentIndex(i)
                    break

        # Config values
        self.ui.start_angle_spin_box.setValue(self.ring.config.start_position)
        self.ui.direction_combo_box.setCurrentIndex(0 if self.ring.config.direction == -1 else 1)
        self.ui.line_width_spin_box.setRange(1, 100)
        self.ui.line_width_spin_box.setValue(self.ring.config.line_width)

        # Colors
        self.ui.ring_color_button.set_color(self.ring.color)
        self.ui.color_sync_button.setCheckable(True)
        self.ui.color_sync_button.setChecked(self.ring.config.link_colors)

        # Set initial button state based on link_colors
        if self.ring.config.link_colors:
            self.ui.color_sync_button.setIcon(material_icon("link"))
            self.ui.color_sync_button.setToolTip(
                "Colors are linked - background derives from main color"
            )
            self.ui.background_color_button.setEnabled(False)
            self.ui.background_color_label.setEnabled(False)
            # Trigger sync to ensure background color is derived from main color
            self.ring.set_color(self.ring.config.color)
            self.ui.background_color_button.set_color(self.ring.background_color)
        else:
            self.ui.color_sync_button.setIcon(material_icon("link_off"))
            self.ui.color_sync_button.setToolTip(
                "Colors are unlinked - set background independently"
            )
            self.ui.background_color_button.setEnabled(True)
            self.ui.background_color_label.setEnabled(True)
            self.ui.background_color_button.set_color(self.ring.background_color)

        self.ui.color_sync_button.toggled.connect(self._toggle_color_link)

    # --------------------------------------------------------

    def _toggle_color_link(self, checked: bool):
        """Toggle the color linking between main and background color"""
        self.ring.config.link_colors = checked

        # Update button icon and tooltip based on state
        if checked:
            self.ui.color_sync_button.setIcon(material_icon("link"))
            self.ui.color_sync_button.setToolTip(
                "Colors are linked - background derives from main color"
            )
            # Trigger background color update by calling set_color
            self.ring.set_color(self.ring.config.color)
            # Update UI to show the new background color
            self.ui.background_color_button.set_color(self.ring.background_color)
        else:
            self.ui.color_sync_button.setIcon(material_icon("link_off"))
            self.ui.color_sync_button.setToolTip(
                "Colors are unlinked - set background independently"
            )

        # Enable/disable background color controls based on link state
        self.ui.background_color_button.setEnabled(not checked)
        self.ui.background_color_label.setEnabled(not checked)

    def _on_ring_color_changed(self, color: QColor):
        """Handle ring color changes and update background if colors are linked"""
        self.ring.set_color(color)
        # If colors are linked, update the background color button to show the new derived color
        if self.ring.config.link_colors:
            self.ui.background_color_button.set_color(self.ring.background_color)

    def _update_min_max(self):
        self.ui.value_spin_box.setRange(self.ui.min_spin_box.value(), self.ui.max_spin_box.value())
        self.ring.set_min_max_values(self.ui.min_spin_box.value(), self.ui.max_spin_box.value())

    def _update_direction(self, index: int):
        self.ring.config.direction = -1 if index == 0 else 1
        self.ring.update()

    @SafeSlot(str)
    def _on_device_changed(self, device: str):
        signal = self.ui.signal_combo_box.get_signal_name()
        self.ring.set_update("device", device=device, signal=signal)
        self.ring.config.device = device

    @SafeSlot(str)
    def _on_signal_changed(self, signal: str):
        device = self.ui.device_combo_box.currentText()
        signal = self.ui.signal_combo_box.get_signal_name()
        if not device or device not in self.ring.bec_dispatcher.client.device_manager.devices:
            return
        self.ring.set_update("device", device=device, signal=signal)
        self.ring.config.signal = signal

    @staticmethod
    def _unify_mode_string(mode: str) -> str:
        """Convert mode string to a unified format"""
        mode = mode.lower()
        if mode == "scan progress":
            return "scan"
        if mode == "device readback":
            return "device"
        return mode

    @staticmethod
    def _get_display_mode_string(mode: str) -> str:
        """Convert mode string to display format"""
        match mode:
            case "manual":
                return "Manual"
            case "scan":
                return "Scan Progress"
            case "device":
                return "Device Readback"
        return mode.capitalize()

    def _update_mode(self, mode: str):
        """Update the ring's mode based on combo box selection"""
        mode = self._unify_mode_string(mode)
        match mode:
            case "manual":
                self.ring.set_update("manual")
            case "scan":
                self.ring.set_update("scan")
            case "device":
                self.ring.set_update("device", device=self.ui.device_combo_box.currentText())
        self._set_widget_mode_enabled(mode)

    def _set_widget_mode_enabled(self, mode: str):
        """Show/hide controls based on the current mode"""
        mode = self._unify_mode_string(mode)
        self.ui.device_combo_box.setEnabled(mode == "device")
        self.ui.signal_combo_box.setEnabled(mode == "device")
        self.ui.device_label.setEnabled(mode == "device")
        self.ui.signal_label.setEnabled(mode == "device")
        self.ui.min_label.setEnabled(mode in ["manual", "device"])
        self.ui.max_label.setEnabled(mode in ["manual", "device"])
        self.ui.value_label.setEnabled(mode == "manual")
        self.ui.value_spin_box.setEnabled(mode == "manual")
        self.ui.min_spin_box.setEnabled(mode in ["manual", "device"])
        self.ui.max_spin_box.setEnabled(mode in ["manual", "device"])

    def _delete_self(self):
        """Delete this ring from the container"""
        if self.ring in self.container.rings:
            self.container.rings.remove(self.ring)
            self.ring.deleteLater()

        self.cleanup()

    def cleanup(self):
        """Cleanup the card widget"""
        self.ui.device_combo_box.close()
        self.ui.device_combo_box.deleteLater()
        self.ui.signal_combo_box.close()
        self.ui.signal_combo_box.deleteLater()
        self.close()
        self.deleteLater()


# ============================================================
# Ring settings widget
# ============================================================


class RingSettings(SettingWidget):
    def __init__(
        self, parent=None, target_widget: RingProgressBar | None = None, popup=False, **kwargs
    ):
        super().__init__(parent=parent, **kwargs)

        self.setProperty("skip_settings", True)
        self.target_widget = target_widget
        self.popup = popup
        if not target_widget:
            return
        self.container: RingProgressContainerWidget = target_widget.ring_progress_bar
        self.original_num_bars = len(self.container.rings)
        self.original_configs = [ring.config.model_dump() for ring in self.container.rings]

        layout = QVBoxLayout(self)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.setSpacing(10)

        add_button = QPushButton(material_icon("add"), "Add Ring")
        add_button.clicked.connect(self.add_ring)

        self.center_label_edit = QLineEdit(self.container.center_label.text())
        self.center_label_edit.setPlaceholderText("Center Label")
        self.center_label_edit.textChanged.connect(self._update_center_label)

        self.colormap_toggle = QPushButton()
        self.colormap_toggle.setCheckable(True)
        self.colormap_toggle.setIcon(material_icon("palette"))
        self.colormap_toggle.setToolTip(
            f"Colormap mode is {'enabled' if self.container.color_map else 'disabled'}"
        )
        self.colormap_toggle.toggled.connect(self._toggle_colormap_mode)

        self.colormap_button = BECColorMapWidget(parent=self)
        self.colormap_button.setToolTip("Set a global colormap for all rings")
        self.colormap_button.colormap_changed_signal.connect(self._set_global_colormap)

        toolbar = QHBoxLayout()

        toolbar.addWidget(add_button)
        toolbar.addWidget(self.center_label_edit)

        toolbar.addStretch()
        toolbar.addWidget(self.colormap_toggle)
        toolbar.addWidget(self.colormap_button)

        layout.addLayout(toolbar)

        self.scroll = QScrollArea()
        self.scroll.setWidgetResizable(True)

        self.cards_container = QWidget()
        self.cards_layout = QVBoxLayout(self.cards_container)
        self.cards_layout.setSpacing(10)
        self.cards_layout.addStretch()

        self.scroll.setWidget(self.cards_container)
        layout.addWidget(self.scroll)

        self.refresh_from_container()
        self.original_label = self.container.center_label.text()

    def sizeHint(self) -> QSize:
        return QSize(720, 520)

    def refresh_from_container(self):
        if not self.container:
            return

        for ring in self.container.rings:
            card = RingCardWidget(ring, self.container)
            self.cards_layout.insertWidget(self.cards_layout.count() - 1, card)

        if self.container.color_map:
            self.colormap_button.colormap = self.container.color_map
        self.colormap_toggle.setChecked(bool(self.container.color_map))

    @SafeSlot()
    def add_ring(self):
        if not self.container:
            return
        self.container.add_ring()
        ring = self.container.rings[len(self.container.rings) - 1]
        if ring:
            card = RingCardWidget(ring, self.container)
            self.cards_layout.insertWidget(self.cards_layout.count() - 1, card)

        # If a global colormap is set, apply it
        if self.container.color_map:
            self._toggle_colormap_mode(bool(self.container.color_map))

    @SafeSlot(str)
    def _update_center_label(self, text: str):
        if not self.container:
            return
        self.container.center_label.setText(text)

    @SafeSlot(bool)
    def _toggle_colormap_mode(self, enabled: bool):
        self.colormap_toggle.setToolTip(f"Colormap mode is {'enabled' if enabled else 'disabled'}")
        if enabled:
            colormap = self.colormap_button.colormap
            self._set_global_colormap(colormap)
        else:
            self.container.color_map = ""
        for i in range(self.cards_layout.count() - 1):  # -1 to exclude the stretch
            widget = self.cards_layout.itemAt(i).widget()
            if not isinstance(widget, RingCardWidget):
                continue
            widget.ui.ring_color_button.setEnabled(not enabled)
            widget.ui.ring_color_button.setToolTip(
                "Disabled in colormap mode" if enabled else "Set the ring color"
            )
            widget.ui.ring_color_label.setEnabled(not enabled)
            widget.ui.background_color_button.setEnabled(
                not enabled and not widget.ring.config.link_colors
            )
            widget.ui.color_sync_button.setEnabled(not enabled)

    @SafeSlot(str)
    def _set_global_colormap(self, colormap: str):
        if not self.container:
            return
        self.container.set_colors_from_map(colormap)

        # Update all ring card color buttons to reflect the new colors
        for i in range(self.cards_layout.count() - 1):  # -1 to exclude the stretch
            widget = self.cards_layout.itemAt(i).widget()
            if not isinstance(widget, RingCardWidget):
                continue
            widget.ui.ring_color_button.set_color(widget.ring.color)
            if widget.ring.config.link_colors:
                widget.ui.background_color_button.set_color(widget.ring.background_color)

    @SafeSlot()
    def accept_changes(self):
        if not self.container:
            return

        self.original_configs = [ring.config.model_dump() for ring in self.container.rings]

        for i, ring in enumerate(self.container.rings):
            ring.setGeometry(self.container.rect())
            ring.gap = self.container.gap * i
            ring.show()  # Ensure ring is visible
            ring.raise_()  # Bring ring to front

        self.container.center_label.setText(self.center_label_edit.text())
        self.original_label = self.container.center_label.text()
        self.original_num_bars = len(self.container.rings)

        self.container.update()

    def cleanup(self):
        """
        Cleanup the settings widget.
        """
        # Remove any rings that were added but not applied
        if not self.container:
            return
        if len(self.container.rings) > self.original_num_bars:
            remove_rings = self.container.rings[self.original_num_bars :]
            for ring in remove_rings:
                self.container.rings.remove(ring)
                ring.deleteLater()
        rings_to_add = max(0, self.original_num_bars - len(self.container.rings))
        for _ in range(rings_to_add):
            self.container.add_ring()

        # apply original configs to all rings
        for i, ring in enumerate(self.container.rings):
            ring.config = ring.config.model_validate(self.original_configs[i])

        for i in range(self.cards_layout.count()):
            item = self.cards_layout.itemAt(i)
            if not item or not item.widget():
                continue
            widget: RingCardWidget = item.widget()
            widget.cleanup()
        self.container.update()
        self.container.center_label.setText(self.original_label)



from __future__ import annotations

import operator
import os
from enum import Enum, auto
from functools import partial, reduce
from glob import glob
from pathlib import Path
from typing import Protocol

import bec_lib
from bec_lib.atlas_models import HashableDevice, HashableDeviceSet
from bec_lib.bec_yaml_loader import yaml_load
from bec_lib.logger import bec_logger
from bec_lib.plugin_helper import plugin_package_name, plugin_repo_path, plugins_installed

logger = bec_logger.logger

# use the last n recovery files
_N_RECOVERY_FILES = 3
_BASE_REPO_PATH = Path(os.path.dirname(bec_lib.__file__)) / "../.."


def get_backend() -> DeviceResourceBackend:
    return _ConfigFileBackend()


class HashModel(str, Enum):
    DEFAULT = auto()
    DEFAULT_DEVICECONFIG = auto()
    DEFAULT_EPICS = auto()


class DeviceResourceBackend(Protocol):
    @property
    def tag_groups(self) -> dict[str, set[HashableDevice]]:
        """A dictionary of all available devices separated by tag groups. The same device may
        appear more than once (in different groups)."""
        ...

    @property
    def all_devices(self) -> set[HashableDevice]:
        """A set of all available devices. The same device may not appear more than once."""
        ...

    @property
    def untagged_devices(self) -> set[HashableDevice]:
        """A set of all untagged devices. The same device may not appear more than once."""
        ...

    @property
    def allowed_sort_keys(self) -> set[str]:
        """A set of all fields which you may group devices by"""
        ...

    def tags(self) -> set[str]:
        """Returns a set of all the tags in all available devices."""
        ...

    def tag_group(self, tag: str) -> set[HashableDevice]:
        """Returns a set of the devices in the tag group with the given key."""
        ...

    def group_by_key(self, key: str) -> dict[str, set[HashableDevice]]:
        """Return a dict of all devices, organised by the specified key, which must be one of
        the string keys in the Device model."""
        ...


def _devices_from_file(file: str, include_source: bool = True):
    data = yaml_load(file, process_includes=False)
    return HashableDeviceSet(
        HashableDevice.model_validate(
            dev | {"name": name, "source_files": {file} if include_source else set()}
        )
        for name, dev in data.items()
    )


class _ConfigFileBackend(DeviceResourceBackend):
    def __init__(self) -> None:
        self._raw_device_set: set[HashableDevice] = self._get_config_from_backup_files()
        if plugins_installed() == 1:
            self._raw_device_set.update(
                self._get_configs_from_plugin_files(
                    Path(plugin_repo_path()) / plugin_package_name() / "device_configs/"
                )
            )
        self._device_groups = self._get_tag_groups()

    def _get_config_from_backup_files(self):
        dir = _BASE_REPO_PATH / "logs/device_configs/recovery_configs"
        files = sorted(glob("*.yaml", root_dir=dir))
        last_n_files = files[-_N_RECOVERY_FILES:]
        return reduce(
            operator.or_,
            map(
                partial(_devices_from_file, include_source=False),
                (str(dir / f) for f in last_n_files),
            ),
            set(),
        )

    def _get_configs_from_plugin_files(self, dir: Path):
        files = glob("*.yaml", root_dir=dir, recursive=True)
        return reduce(operator.or_, map(_devices_from_file, (str(dir / f) for f in files)), set())

    def _get_tag_groups(self) -> dict[str, set[HashableDevice]]:
        return {
            tag: set(filter(lambda dev: tag in dev.deviceTags, self._raw_device_set))
            for tag in self.tags()
        }

    @property
    def tag_groups(self):
        return self._device_groups

    @property
    def all_devices(self):
        return self._raw_device_set

    @property
    def untagged_devices(self):
        return {d for d in self._raw_device_set if d.deviceTags == set()}

    @property
    def allowed_sort_keys(self) -> set[str]:
        return {n for n, info in HashableDevice.model_fields.items() if info.annotation is str}

    def tags(self) -> set[str]:
        return reduce(operator.or_, (dev.deviceTags for dev in self._raw_device_set), set())

    def tag_group(self, tag: str) -> set[HashableDevice]:
        return self.tag_groups[tag]

    def group_by_key(self, key: str) -> dict[str, set[HashableDevice]]:
        if key not in self.allowed_sort_keys:
            raise ValueError(f"Cannot group available devices by model key {key}")
        group_names: set[str] = {getattr(item, key) for item in self._raw_device_set}
        return {g: {d for d in self._raw_device_set if getattr(d, key) == g} for g in group_names}



"""Module for the device configuration form widget for EpicsMotor, EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV"""

from copy import deepcopy
from typing import Any, Type

from bec_lib.atlas_models import Device as DeviceModel
from bec_lib.logger import bec_logger
from ophyd_devices.interfaces.device_config_templates.ophyd_templates import OPHYD_DEVICE_TEMPLATES
from pydantic import BaseModel
from pydantic_core import PydanticUndefinedType
from qtpy import QtWidgets

from bec_widgets.widgets.control.device_manager.components.device_config_template.template_items import (
    DEVICE_CONFIG_FIELDS,
    DEVICE_FIELDS,
    DeviceConfigField,
    DeviceTagsWidget,
    InputLineEdit,
    LimitInputWidget,
    OnFailureComboBox,
    ParameterValueWidget,
    ReadoutPriorityComboBox,
)
from bec_widgets.widgets.utility.toggle.toggle import ToggleSwitch

logger = bec_logger.logger


class DeviceConfigTemplate(QtWidgets.QWidget):
    """
    Device Configuration Template Widget.
    Current supported templates follow the structure in
    ophyd_devices.interfaces.device_config_templates.ophyd_templates.OPHYD_DEVICE_TEMPLATES.

    Args:
        parent (QtWidgets.QWidget, optional)         : Parent widget. Defaults to None.
        client (BECClient, optional)                 : BECClient instance. Defaults to None.
        template (dict[str, any], optional)          : Device configuration template. If None,
                                                       the "CustomDevice" template will be used. Defaults to None.
    """

    RPC = False

    def __init__(self, parent=None, template: dict[str, any] = None):
        super().__init__(parent=parent)
        if template is None:
            template = OPHYD_DEVICE_TEMPLATES["CustomDevice"]["CustomDevice"]
        self.template = template
        self._device_fields = deepcopy(DEVICE_FIELDS)
        self._device_config_fields = deepcopy(DEVICE_CONFIG_FIELDS)
        self._unknown_device_config_entry: dict[str, any] = {}

        # Dict to store references to input widgets
        self._widgets: dict[str, QtWidgets.QWidget] = {}

        # Two column layout
        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(2, 0, 2, 0)
        layout.setSpacing(2)
        self.setLayout(layout)

        # Left hand side, settings, connection and advanced settings
        self._left_layout = QtWidgets.QVBoxLayout()
        self._left_layout.setContentsMargins(2, 2, 2, 2)
        self._left_layout.setSpacing(4)
        # Settings box, name | deviceClass | description
        self.settings_box = self._create_settings_box()
        # Device Config settings box | dynamic fields from deviceConfig
        self.connection_settings_box = self._create_connection_settings_box()
        # Advanced Control box | readoutPriority | onFailure | softwareTrigger | enabled | readOnly
        self.advanced_control_box = self._create_advanced_control_box()
        # Add boxes to left layout
        self._left_layout.addWidget(self.settings_box)
        self._left_layout.addWidget(self.connection_settings_box)
        self._left_layout.addWidget(self.advanced_control_box)
        layout.addLayout(self._left_layout)

        # Right hand side, advanced settings
        self._right_layout = QtWidgets.QVBoxLayout()
        self._right_layout.setContentsMargins(2, 2, 2, 2)
        self._right_layout.setSpacing(4)
        layout.addLayout(self._right_layout)
        # Create Additional Settings box
        self.additional_settings_box = self.create_additional_settings()
        self._right_layout.addWidget(self.additional_settings_box)

        # Set default values
        self.reset_to_defaults()

    def _clear_layout(self, layout: QtWidgets.QLayout) -> None:
        """Clear a layout recursively. If the layout contains sub-layouts, they will also be cleared."""
        while layout.count():
            item = layout.takeAt(0)
            if item.widget():
                item.widget().close()
                item.widget().deleteLater()
            if item.layout():
                self._clear_layout(item.layout())

    def reset_to_defaults(self) -> None:
        """Reset all fields to default values."""
        self._widgets.pop("deviceConfig", None)
        self._clear_layout(self.connection_settings_box.layout())

        # Recreate Connection Settings box
        layout: QtWidgets.QGridLayout = self.connection_settings_box.layout()
        self._fill_connection_settings_box(self.connection_settings_box, layout)

        # Reset Settings and Advanced Control boxes
        for field_name, widget in self._widgets.items():
            if field_name in self.template:
                self._set_value_for_widget(widget, self.template[field_name])
            else:
                self._set_default_entry(field_name, widget)

    def change_template(self, template: dict[str, any]) -> None:
        """
        Change the template and update the form fields accordingly.

        Args:
            template (dict[str, any]): New device configuration template.
        """
        self.template = template
        self.reset_to_defaults()

    def get_config_fields(self) -> dict:
        """Retrieve the current configuration from the input fields."""
        config: dict[str, any] = {}
        for device_entry, widget in self._widgets.items():
            config[device_entry] = self._get_entry_for_widget(widget)
        if self._unknown_device_config_entry:
            if "deviceConfig" not in config:
                config["deviceConfig"] = {}
            config["deviceConfig"].update(self._unknown_device_config_entry)
        return config

    def set_config_fields(self, config: dict) -> None:
        """
        Set the configuration fields based on the provided config dictionary.

        Args:
            config (dict): Configuration dictionary to set the fields.
        """
        # Clear storage for unknown entries
        self._unknown_device_config_entry.clear()
        if self.template.get("deviceClass", "") != config.get("deviceClass", ""):
            logger.warning(
                f"Device class {config.get('deviceClass', '')} does not match template device class {self.template.get('deviceClass', '')}. Using custom device template."
            )
            self.change_template(OPHYD_DEVICE_TEMPLATES["CustomDevice"]["CustomDevice"])
        else:
            self.reset_to_defaults()
        self._fill_fields_from_config(config)

    def _fill_fields_from_config(self, model: dict) -> None:
        """
        Fill the form fields base on the provided configuration dictionary.
        Please note, deviceConfig is handled separately through _fill_connection_settings_box
        as this depends on the template used.

        Args:
            model (dict): Configuration dictionary to fill the fields.
        """
        for key, value in model.items():
            if key == "name":
                wid = self._widgets["name"]
                wid.setText(value or "")
            elif key == "deviceClass":
                wid = self._widgets["deviceClass"]
                wid.setText(value or "")
                if "deviceClass" in self.template:
                    wid.setEnabled(False)
                else:
                    wid.setEnabled(True)
            elif key == "deviceConfig" and isinstance(
                self._widgets.get("deviceConfig", None), dict
            ):
                # If _widgets["deviceConfig"] is a dict, we have individual widgets for each field
                for sub_key, sub_value in value.items():
                    widget = self._widgets["deviceConfig"].get(sub_key, None)
                    if widget is None:
                        logger.warning(
                            f"Widget for key {sub_key} not found in deviceConfig widgets."
                        )
                        # Store any unknown entry fields
                        self._unknown_device_config_entry[sub_key] = sub_value
                        continue
                    self._set_value_for_widget(widget, sub_value)
            else:
                widget = self._widgets.get(key, None)
                if widget is not None:
                    self._set_value_for_widget(widget, value)

    def _set_value_for_widget(self, widget: QtWidgets.QWidget, value: Any) -> None:
        """
        Set the value for a widget based on its type.

        Args:
            widget (QtWidgets.QWidget): The widget to set the value for.
            value (any): The value to set.
        """
        if isinstance(widget, (ParameterValueWidget)) and isinstance(value, dict):
            for param, val in value.items():
                widget.add_parameter_line(param, val)
        elif isinstance(widget, DeviceTagsWidget) and isinstance(value, (list, tuple, set)):
            for tag in value:
                widget.add_parameter_line(tag or "")
        elif isinstance(widget, InputLineEdit):
            widget.setText(str(value or ""))
        elif isinstance(widget, ToggleSwitch):
            widget.setChecked(bool(value))
        elif isinstance(widget, LimitInputWidget):
            widget.set_limits(value)
        elif isinstance(widget, QtWidgets.QComboBox):
            index = widget.findText(value)
            if index != -1:
                widget.setCurrentIndex(index)
        elif isinstance(widget, QtWidgets.QTextEdit):
            widget.setPlainText(str(value or ""))
        else:
            logger.warning(f"Unsupported widget type for setting value: {type(widget)}")

    def _get_entry_for_widget(self, widget: QtWidgets.QWidget) -> any:
        """
        Get the value from a widget based on its type.

        Args:
            widget (QtWidgets.QWidget): The widget to get the value from.
        Returns:
            any: The value retrieved from the widget.
        """
        if isinstance(widget, (ParameterValueWidget, DeviceTagsWidget)):
            return widget.parameters()
        elif isinstance(widget, InputLineEdit):
            return widget.text().strip()
        elif isinstance(widget, ToggleSwitch):
            return widget.isChecked()
        elif isinstance(widget, LimitInputWidget):
            return widget.get_limits()
        elif isinstance(widget, QtWidgets.QComboBox):
            return widget.currentText()
        elif isinstance(widget, QtWidgets.QTextEdit):
            return widget.toPlainText()
        elif isinstance(widget, dict):
            result = {}
            for sub_entry, sub_widget in widget.items():
                result[sub_entry] = self._get_entry_for_widget(sub_widget)
            return result
        else:
            logger.warning(f"Unsupported widget type for getting entry: {type(widget)}")
            return None

    def _create_device_field(
        self, field_name: str, field_info: DeviceConfigField | None = None
    ) -> tuple[QtWidgets.QLabel, QtWidgets.QWidget]:
        """
        Create a device field based on the field name. If field_info is not provided,
        a default label and input widget will be created.

        Args:
            field_name (str): Name of the field.
            field_info (DeviceConfigField | None, optional): Information about the field. Defaults to None.
        """
        if field_info is None:
            label = QtWidgets.QLabel(field_name, parent=self)
            input_widget = QtWidgets.QLineEdit(parent=self)
            return label, input_widget

        label_text = field_info.label
        label = QtWidgets.QLabel(label_text, parent=self)
        if field_info.required:
            label_text = label.text()
            label_text += " *"
            label.setText(label_text)
            label.setStyleSheet("font-weight: bold;")
        input_widget = field_info.widget_cls(parent=self)
        if field_info.placeholder_text:
            if hasattr(input_widget, "setPlaceholderText"):
                input_widget.setPlaceholderText(field_info.placeholder_text)
        if field_info.static:
            input_widget.setEnabled(False)
        if field_info.validation_callback:
            # Attach validation callback if provided
            if isinstance(input_widget, InputLineEdit):
                input_widget: InputLineEdit
                for callback in field_info.validation_callback:
                    input_widget.register_validation_callback(callback)
        if field_info.default is not None:
            # Set default value
            if isinstance(input_widget, QtWidgets.QLineEdit):
                input_widget.setText(str(field_info.default))
            elif isinstance(input_widget, QtWidgets.QTextEdit):
                input_widget.setPlainText(str(field_info.default))
            elif isinstance(input_widget, ToggleSwitch):
                input_widget.setChecked(bool(field_info.default))
            elif isinstance(input_widget, (ReadoutPriorityComboBox, OnFailureComboBox)):
                index = input_widget.findText(field_info.default)
                if index != -1:
                    input_widget.setCurrentIndex(index)
        return label, input_widget

    def _create_group_box_with_grid_layout(
        self, title: str
    ) -> tuple[QtWidgets.QGroupBox, QtWidgets.QGridLayout]:
        """Create a group box with a grid layout."""
        box = QtWidgets.QGroupBox(title)
        layout = QtWidgets.QGridLayout(box)
        layout.setContentsMargins(4, 8, 4, 8)
        layout.setSpacing(4)
        box.setLayout(layout)
        return box, layout

    def _set_default_entry(self, field_name: str, widget: QtWidgets.QWidget) -> None:
        """
        Set the default value for a given field in the form based on the Pydantic model.

        Args:
            field_name (str): Name of the field.
            widget (QtWidgets.QWidget): The widget to set the default value for.
        """
        if field_name == "enabled":
            widget.setChecked(True)
            return
        if field_name == "readOnly":
            widget.setChecked(False)
            return
        default = self._get_default_for_device_config_field(field_name) or ""
        widget.setEnabled(True)
        if isinstance(widget, QtWidgets.QComboBox):
            index = widget.findText(default)
            if index != -1:
                widget.setCurrentIndex(index)
        elif isinstance(widget, (QtWidgets.QTextEdit, QtWidgets.QLineEdit)):
            widget.setText(str(default))
        elif isinstance(widget, ToggleSwitch):
            widget.setChecked(bool(default))
        elif isinstance(widget, (ParameterValueWidget, DeviceTagsWidget)):
            widget.clear_widget()

    def _get_default_for_device_config_field(self, field_name: str) -> any:
        """
        Get the default value for a given deviceConfig field based on the Pydantic model.

        Args:
            field_name (str): Name of the deviceConfig field.
        Returns:
            any: The default value for the field, or None if not found.
        """
        model_properties: dict = DeviceModel.model_json_schema()["properties"]
        if field_name in model_properties:
            field_info = model_properties[field_name]
            default = field_info.get("default", None)
            if default:
                return default
        return None

    ### Box creation methods ###

    def _create_box(self, box_title: str, field_names: list[str]) -> QtWidgets.QGroupBox:
        """
        Create a box layout with specific fields. If field_names are in _device_fields,
        their corresponding widgets will be used.
        """
        # Create box
        box, layout = self._create_group_box_with_grid_layout(box_title)
        box.setLayout(layout)

        for ii, field_name in enumerate(field_names):
            label, input_widget = self._create_device_field(
                field_name, self._device_fields.get(field_name, None)
            )
            layout.addWidget(label, ii, 0)
            layout.addWidget(input_widget, ii, 1)
            self._widgets[field_name] = input_widget
        return box

    def _create_settings_box(self) -> QtWidgets.QGroupBox:
        """Create the settings box widget."""
        box = self._create_box("Settings", ["name", "deviceClass", "description"])
        layout = box.layout()
        # Set column stretch
        layout.setColumnStretch(0, 0)
        layout.setColumnStretch(1, 1)
        return box

    def _create_advanced_control_box(self) -> QtWidgets.QGroupBox:
        """Create the advanced control box widget."""
        # Set up advanced control box
        box = self._create_box("Advanced Control", ["readoutPriority", "onFailure"])
        layout = box.layout()
        for ii, field_name in enumerate(["enabled", "readOnly", "softwareTrigger"]):
            label, input_widget = self._create_device_field(
                field_name, self._device_fields.get(field_name, None)
            )
            layout.addWidget(label, ii, 2)
            layout.addWidget(input_widget, ii, 3)
            self._widgets[field_name] = input_widget
        return box

    def _create_connection_settings_box(self) -> QtWidgets.QGroupBox:
        """Create the connection settings box widget. These are all entries in the deviceConfig field."""
        box, layout = self._create_group_box_with_grid_layout("Connection Settings")
        box = self._fill_connection_settings_box(box, layout)
        return box

    def _fill_connection_settings_box(
        self, box: QtWidgets.QGroupBox, layout: QtWidgets.QGridLayout
    ) -> QtWidgets.QGroupBox:
        """Fill the connection settings box based on the deviceConfig template."""
        if not self.template.get("deviceConfig", {}):
            widget = ParameterValueWidget(parent=self)
            widget.setToolTip(
                "Add custom deviceConfig entries as key-value pairs in the tree view."
            )
            layout.addWidget(widget, 0, 0)
            self._widgets["deviceConfig"] = widget
            return box
        # If template specifies deviceConfig fields, create them
        self._widgets["deviceConfig"] = {}
        model: Type[BaseModel] = self.template["deviceConfig"]
        for field_name, field in model.model_fields.items():
            field_info = self._device_config_fields.get(field_name, None)
            default = field.get_default()
            if isinstance(default, PydanticUndefinedType):
                default = None
            if field_info:
                if field.is_required():
                    field_info.required = True
                if field.description:
                    field_info.placeholder_text = field.description
                if default is not None:
                    field_info.default = default
            label, input_widget = self._create_device_field(field_name, field_info)
            row = layout.rowCount()
            layout.addWidget(label, row, 0)
            layout.addWidget(input_widget, row, 1)
            self._widgets["deviceConfig"][field_name] = input_widget
        return box

    def create_additional_settings(self) -> QtWidgets.QGroupBox:
        """Create the additional settings box widget."""
        box, layout = self._create_group_box_with_grid_layout("Additional Settings")
        toolbox = QtWidgets.QToolBox(parent=self)
        layout.addWidget(toolbox, 0, 0)
        user_parameters_widget = ParameterValueWidget(parent=self)
        self._widgets["userParameter"] = user_parameters_widget
        toolbox.addItem(user_parameters_widget, "User Parameter")
        device_tags_widget = DeviceTagsWidget(parent=self)
        toolbox.addItem(device_tags_widget, "Device Tags")
        toolbox.setCurrentIndex(1)
        self._widgets["deviceTags"] = device_tags_widget
        return box


if __name__ == """__main__""":  # pragma: no cover
    import sys

    app = QtWidgets.QApplication(sys.argv)
    import yaml
    from bec_qthemes import apply_theme

    from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton

    apply_theme("light")

    class TestWidget(QtWidgets.QWidget):
        pass

    w = TestWidget()
    w_layout = QtWidgets.QVBoxLayout(w)
    w_layout.setContentsMargins(0, 0, 0, 0)
    w_layout.setSpacing(20)
    dark_mode_button = DarkModeButton()
    w_layout.addWidget(dark_mode_button)
    test_motor = "EpicsMotor"
    config_form = DeviceConfigTemplate(template=OPHYD_DEVICE_TEMPLATES[test_motor][test_motor])
    w_layout.addWidget(config_form)
    button_layout = QtWidgets.QHBoxLayout()
    button = QtWidgets.QPushButton("Get Config")
    button.clicked.connect(
        lambda: print("Device Config:", yaml.dump(config_form.get_config_fields(), indent=4))
    )
    button_layout.addWidget(button)
    button2 = QtWidgets.QPushButton("Reset")
    button2.clicked.connect(config_form.reset_to_defaults)
    button_layout.addWidget(button2)
    combo = QtWidgets.QComboBox()
    combo_keys = [
        "EpicsMotor",
        "EpicsSignal",
        "EpicsSignalRO",
        "EpicsSignalWithRBV",
        "CustomDevice",
    ]
    combo.addItems(combo_keys)
    combo.setCurrentText(test_motor)

    def text_changed(text: str) -> None:
        if text.startswith("EpicsMotor"):
            if text == "EpicsMotor":
                template = OPHYD_DEVICE_TEMPLATES[text][text]
            else:
                template = OPHYD_DEVICE_TEMPLATES["EpicsMotor"][text]
        elif text.startswith("EpicsSignal"):
            if text == "EpicsSignal":
                template = OPHYD_DEVICE_TEMPLATES[text][text]
            else:
                template = OPHYD_DEVICE_TEMPLATES["EpicsSignal"][text]
        else:
            template = OPHYD_DEVICE_TEMPLATES["CustomDevice"]["CustomDevice"]
        config_form.change_template(template)

    combo.currentTextChanged.connect(text_changed)
    button_layout.addWidget(button)
    button_layout.addWidget(combo)
    w_layout.addLayout(button_layout)
    w.resize(1200, 600)
    w.show()
    sys.exit(app.exec_())



from __future__ import annotations

from decimal import Decimal

from bec_lib.logger import bec_logger
from bec_lib.metadata_schema import get_metadata_schema_for_scan
from pydantic import Field
from qtpy.QtWidgets import QApplication, QComboBox, QHBoxLayout, QVBoxLayout, QWidget

from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.utils.expandable_frame import ExpandableGroupFrame
from bec_widgets.utils.forms_from_types.forms import PydanticModelForm
from bec_widgets.widgets.editors.dict_backed_table import DictBackedTable

logger = bec_logger.logger


class ScanMetadata(PydanticModelForm):

    RPC = False

    def __init__(
        self,
        parent=None,
        client=None,
        scan_name: str | None = None,
        initial_extras: list[list[str]] | None = None,
        **kwargs,
    ):
        """Dynamically generates a form for inclusion of metadata for a scan. Uses the
        metadata schema registry supplied in the plugin repo to find pydantic models
        associated with the scan type. Sets limits for numerical values if specified.

        Args:
            scan_name (str): The scan for which to generate a metadata form
            Initial_extras (list[list[str]]): Initial data with which to populate the additional
                                              metadata table - inner lists should be key-value pairs
        """

        # self.populate() gets called in super().__init__
        # so make sure self._additional_metadata exists
        self._additional_md_box = ExpandableGroupFrame(
            parent, "Additional metadata", expanded=False
        )
        self._additional_md_box_layout = QHBoxLayout()
        self._additional_md_box.set_layout(self._additional_md_box_layout)

        self._additional_metadata = DictBackedTable(parent, initial_extras or [])
        self._scan_name = scan_name or ""
        self._md_schema = get_metadata_schema_for_scan(self._scan_name)
        self._additional_metadata.data_changed.connect(self.validate_form)
        self._hide_optional_metadata = False

        super().__init__(parent=parent, data_model=self._md_schema, client=client, **kwargs)

        self._layout.setContentsMargins(0, 0, 0, 0)
        self._form_grid_container.layout().setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._additional_md_box)
        self._additional_md_box_layout.addWidget(self._additional_metadata)

    @SafeSlot(str)
    def update_with_new_scan(self, scan_name: str):
        self.set_schema_from_scan(scan_name)
        self.validate_form()

    @SafeProperty(bool)
    def hide_optional_metadata(self):  # type: ignore
        """Property to hide the optional metadata table."""
        return self._hide_optional_metadata

    @hide_optional_metadata.setter
    def hide_optional_metadata(self, hide: bool):
        """Setter for the hide_optional_metadata property.

        Args:
            hide(bool): Hide or show the optional metadata table.
        """
        self._hide_optional_metadata = hide
        self._additional_md_box.setVisible(not hide)

    def get_form_data(self):
        """Get the entered metadata as a dict"""
        form_data = self._additional_metadata.dump_dict() | self._dict_from_grid()

        # If scan_name is empty, set it to the current scan
        if "scan_name" in form_data and not form_data["scan_name"]:
            form_data["scan_name"] = self._scan_name

        return form_data

    def populate(self):
        self._additional_metadata.update_disallowed_keys(list(self._md_schema.model_fields.keys()))
        super().populate()

        # Set scan_name field to current scan if it exists and is empty
        if "scan_name" not in self.widget_dict:
            return
        scan_name_widget = self.widget_dict["scan_name"]
        if not hasattr(scan_name_widget, "getValue") or scan_name_widget.getValue():
            return
        if hasattr(scan_name_widget, "setValue"):
            scan_name_widget.setValue(self._scan_name)

    def set_schema_from_scan(self, scan_name: str | None):
        self._scan_name = scan_name or ""
        self.set_schema(get_metadata_schema_for_scan(self._scan_name))


if __name__ == "__main__":  # pragma: no cover
    # pylint: disable=redefined-outer-name
    # pylint: disable=protected-access
    # pylint: disable=disallowed-name

    from unittest.mock import patch

    from bec_lib.metadata_schema import BasicScanMetadata

    from bec_widgets.utils.colors import apply_theme

    class ExampleSchema1(BasicScanMetadata):
        abc: int = Field(gt=0, lt=2000, description="Heating temperature abc", title="A B C")
        foo: str = Field(max_length=12, description="Sample database code", default="DEF123")
        xyz: Decimal = Field(decimal_places=4)
        baz: bool

    class ExampleSchema2(BasicScanMetadata):
        checkbox_up_top: bool
        checkbox_again: bool = Field(
            title="Checkbox Again", description="this one defaults to True", default=True
        )
        different_items: int | None = Field(
            None, description="This is just one different item...", gt=-100, lt=0
        )
        length_limited_string: str = Field(max_length=32)
        float_with_2dp: Decimal = Field(decimal_places=2)

    class ExampleSchema3(BasicScanMetadata):
        optional_with_regex: str | None = Field(None, pattern=r"^\d+-\d+$")

    with patch(
        "bec_lib.metadata_schema._get_metadata_schema_registry",
        lambda: {"scan1": ExampleSchema1, "scan2": ExampleSchema2, "scan3": ExampleSchema3},
    ):
        app = QApplication([])
        w = QWidget()
        selection = QComboBox()
        selection.addItems(["grid_scan", "scan1", "scan2", "scan3"])

        layout = QVBoxLayout()
        w.setLayout(layout)

        scan_metadata = ScanMetadata(
            parent=w,
            scan_name="grid_scan",
            initial_extras=[["key1", "value1"], ["key2", "value2"], ["key3", "value3"]],
        )
        selection.currentTextChanged.connect(scan_metadata.update_with_new_scan)

        layout.addWidget(selection)
        layout.addWidget(scan_metadata)

        apply_theme("dark")
        window = w
        window.show()
        app.exec()



from __future__ import annotations

from bec_qthemes import material_icon
from qtpy.QtCore import QMimeData, Qt, Signal
from qtpy.QtGui import QDrag
from qtpy.QtWidgets import QHBoxLayout, QPushButton, QSizePolicy, QToolButton, QVBoxLayout, QWidget

from bec_widgets.utils.error_popups import SafeProperty


class CollapsibleSection(QWidget):
    """A widget that combines a header button with any content widget for collapsible sections

    This widget contains a header button with a title and a content widget.
    The content widget can be any QWidget. The header button can be expanded or collapsed.
    The header also contains an "Add" button that is only visible when hovering over the section.

    Signals:
        section_reorder_requested(str, str): Emitted when the section is dragged and dropped
                                             onto another section for reordering.
                                             Arguments are (source_title, target_title).
    """

    section_reorder_requested = Signal(str, str)  # (source_title, target_title)

    def __init__(
        self,
        parent=None,
        title="",
        indentation=10,
        show_add_button=False,
        tooltip: str | None = None,
    ):
        super().__init__(parent=parent)
        self.title = title
        self.content_widget = None
        self.setAcceptDrops(True)
        self._expanded = True

        # Setup layout
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setContentsMargins(indentation, 0, 0, 0)
        self.main_layout.setSpacing(0)

        header_layout = QHBoxLayout()
        header_layout.setContentsMargins(0, 0, 4, 0)
        header_layout.setSpacing(0)

        # Create header button
        self.header_button = QPushButton()
        # Apply theme variant for title styling
        self.header_button.setProperty("variant", "title")
        self.header_button.clicked.connect(self.toggle_expanded)

        # Enable drag and drop for reordering
        self.header_button.setAcceptDrops(True)
        self.header_button.mousePressEvent = self._header_mouse_press_event
        self.header_button.mouseMoveEvent = self._header_mouse_move_event
        self.header_button.dragEnterEvent = self._header_drag_enter_event
        self.header_button.dropEvent = self._header_drop_event
        if tooltip:
            self.header_button.setToolTip(tooltip)

        self.drag_start_position = None

        # Add header to layout
        header_layout.addWidget(self.header_button)
        header_layout.addStretch()

        # Add button in header (icon-only)
        self.header_add_button = QToolButton()
        self.header_add_button.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.header_add_button.setFixedSize(28, 28)
        self.header_add_button.setToolTip("Add item")
        self.header_add_button.setVisible(show_add_button)
        self.header_add_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        self.header_add_button.setAutoRaise(True)

        self.header_add_button.setIcon(material_icon("add", size=(28, 28), convert_to_pixmap=False))
        header_layout.addWidget(self.header_add_button)

        self.main_layout.addLayout(header_layout)

        self._update_expanded_state()

    def set_widget(self, widget):
        """Set the content widget for this collapsible section"""
        # Remove existing content widget if any
        if self.content_widget and self.content_widget.parent() == self:
            self.main_layout.removeWidget(self.content_widget)
            self.content_widget.close()
            self.content_widget.deleteLater()

        self.content_widget = widget
        if self.content_widget:
            self.main_layout.addWidget(self.content_widget)

        self._update_expanded_state()

    def _update_appearance(self):
        """Update the header button appearance based on expanded state"""
        # Use material icons with consistent sizing to match tree items
        icon_name = "keyboard_arrow_down" if self.expanded else "keyboard_arrow_right"
        icon = material_icon(icon_name=icon_name, size=(20, 20), convert_to_pixmap=False)

        self.header_button.setIcon(icon)
        self.header_button.setText(self.title)

    def toggle_expanded(self):
        """Toggle the expanded state and update size policy"""
        self.expanded = not self.expanded
        self._update_expanded_state()

    def _update_expanded_state(self):
        """Update the expanded state based on current state"""
        self._update_appearance()
        if self.expanded:
            if self.content_widget:
                self.content_widget.show()
            self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        else:
            if self.content_widget:
                self.content_widget.hide()
            self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)

    @SafeProperty(bool)
    def expanded(self) -> bool:
        """Get the expanded state"""
        return self._expanded

    @expanded.setter
    def expanded(self, value: bool):
        """Set the expanded state programmatically"""
        if not isinstance(value, bool):
            raise ValueError("Expanded state must be a boolean")
        if self._expanded == value:
            return
        self._expanded = value
        self._update_appearance()

    def connect_add_button(self, slot):
        """Connect a slot to the add button's clicked signal.

        Args:
            slot: The function to call when the add button is clicked.
        """
        self.header_add_button.clicked.connect(slot)

    def _header_mouse_press_event(self, event):
        """Handle mouse press on header for drag start"""
        if event.button() == Qt.MouseButton.LeftButton:
            self.drag_start_position = event.pos()
        QPushButton.mousePressEvent(self.header_button, event)

    def _header_mouse_move_event(self, event):
        """Handle mouse move to start drag operation"""
        if event.buttons() & Qt.MouseButton.LeftButton and self.drag_start_position is not None:

            # Check if we've moved far enough to start a drag
            if (event.pos() - self.drag_start_position).manhattanLength() >= 10:

                self._start_drag()
        QPushButton.mouseMoveEvent(self.header_button, event)

    def _start_drag(self):
        """Start the drag operation with a properly aligned widget pixmap"""
        drag = QDrag(self.header_button)
        mime_data = QMimeData()
        mime_data.setText(f"section:{self.title}")
        drag.setMimeData(mime_data)

        # Grab a pixmap of the widget
        widget_pixmap = self.header_button.grab()

        drag.setPixmap(widget_pixmap)

        # Set the hotspot to where the mouse was pressed on the widget
        drag.setHotSpot(self.drag_start_position)

        drag.exec_(Qt.MoveAction)

    def _header_drag_enter_event(self, event):
        """Handle drag enter on header"""
        if event.mimeData().hasText() and event.mimeData().text().startswith("section:"):
            event.acceptProposedAction()
        else:
            event.ignore()

    def _header_drop_event(self, event):
        """Handle drop on header"""
        if event.mimeData().hasText() and event.mimeData().text().startswith("section:"):
            source_title = event.mimeData().text().replace("section:", "")
            if source_title != self.title:
                # Emit signal to parent to handle reordering
                self.section_reorder_requested.emit(source_title, self.title)
            event.acceptProposedAction()
        else:
            event.ignore()



"""
This module provides a utility class for counting and reporting frames per second (FPS) in a PyQtGraph application.

Classes:
    FPSCounter: A class that monitors the paint events of a `ViewBox` to calculate and emit FPS values.

Usage:
    The `FPSCounter` class can be used to monitor the rendering performance of a `ViewBox` in a PyQtGraph application.
    It connects to the `ViewBox`'s paint event and calculates the FPS over a specified interval, emitting the FPS value
    at regular intervals.

Example:
    from qtpy import QtWidgets, QtCore
    import pyqtgraph as pg
    from fps_counter import FPSCounter

    app = pg.mkQApp("FPS Counter Example")
    win = pg.GraphicsLayoutWidget()
    win.show()

    vb = pg.ViewBox()
    plot_item = pg.PlotItem(viewBox=vb)
    win.addItem(plot_item)

    fps_counter = FPSCounter(vb)
    fps_counter.sigFpsUpdate.connect(lambda fps: print(f"FPS: {fps:.2f}"))

    sys.exit(app.exec_())
"""

from time import perf_counter

import pyqtgraph as pg
from qtpy import QtCore


class FPSCounter(QtCore.QObject):
    """
    A utility class for counting and reporting frames per second (FPS).

    This class connects to a `ViewBox`'s paint event to count the number of
    frames rendered and calculates the FPS over a specified interval. It emits
    a signal with the FPS value at regular intervals.

    Attributes:
        sigFpsUpdate (QtCore.Signal): Signal emitted with the FPS value.
        view_box (pg.ViewBox): The `ViewBox` instance to monitor.
    """

    sigFpsUpdate = QtCore.Signal(float)

    def __init__(self, view_box):
        super().__init__()
        self.view_box = view_box
        self.view_box.sigPaint.connect(self.increment_count)
        self.count = 0
        self.last_update = perf_counter()
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.calculate_fps)
        self.timer.start(1000)

    def increment_count(self):
        """
        Increment the frame count when the `ViewBox` is painted.
        """
        self.count += 1

    def calculate_fps(self):
        """
        Calculate the frames per second (FPS) based on the number of frames
        """
        now = perf_counter()
        elapsed = now - self.last_update
        fps = self.count / elapsed if elapsed > 0 else 0.0
        self.last_update = now
        self.count = 0
        self.sigFpsUpdate.emit(fps)

    def cleanup(self):
        """
        Clean up the FPS counter by stopping the timer and disconnecting the signal.
        """
        self.timer.stop()
        self.timer.timeout.disconnect(self.calculate_fps)



from qtpy import QtCore, QtWidgets

from bec_widgets.utils.bec_widget import BECWidget, ConnectionConfig
from bec_widgets.widgets.services.scan_history_browser.components import (
    ScanHistoryDeviceViewer,
    ScanHistoryMetadataViewer,
    ScanHistoryView,
)


class ScanHistoryBrowser(BECWidget, QtWidgets.QWidget):
    """
    ScanHistoryBrowser is a widget combining the scan history view, metadata viewer, and device viewer.

    Target is to provide a popup view for the Waveform Widget to browse the scan history.
    """

    RPC = False
    PLUGIN = False

    def __init__(
        self,
        parent: QtWidgets.QWidget | None = None,
        client=None,
        config: ConnectionConfig = None,
        gui_id: str | None = None,
        theme_update: bool = False,
        **kwargs,
    ):
        """
        Initialize the ScanHistoryBrowser widget.

        Args:
            parent (QtWidgets.QWidget, optional): The parent widget.
            client: The BEC client.
            config (ConnectionConfig, optional): The connection configuration.
            gui_id (str, optional): The GUI ID.
            theme_update (bool, optional): Whether to subscribe to theme updates. Defaults to False.
        """
        super().__init__(
            parent=parent,
            client=client,
            config=config,
            gui_id=gui_id,
            theme_update=theme_update,
            **kwargs,
        )
        layout = QtWidgets.QHBoxLayout()
        self.setLayout(layout)

        self.scan_history_view = ScanHistoryView(
            parent=self, client=client, config=config, gui_id=gui_id, theme_update=theme_update
        )
        self.scan_history_metadata_viewer = ScanHistoryMetadataViewer(
            parent=self, client=client, config=config, gui_id=gui_id, theme_update=theme_update
        )
        self.scan_history_device_viewer = ScanHistoryDeviceViewer(
            parent=self, client=client, config=config, gui_id=gui_id, theme_update=theme_update
        )

        self.init_layout()
        self.connect_signals()

    def init_layout(self):
        """Initialize compact layout for the widget."""
        # Add Scan history view
        layout: QtWidgets.QHBoxLayout = self.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.scan_history_view)
        # Add metadata and device viewers in a vertical layout
        widget = QtWidgets.QWidget(self)
        vertical_layout = QtWidgets.QVBoxLayout()
        vertical_layout.setContentsMargins(0, 0, 0, 0)
        vertical_layout.setSpacing(0)
        vertical_layout.addWidget(self.scan_history_metadata_viewer)
        vertical_layout.addWidget(self.scan_history_device_viewer)
        widget.setLayout(vertical_layout)
        # Add the vertical layout widget to the main layout
        layout.addWidget(widget)

    def connect_signals(self):
        """Connect signals from scan history components."""
        self.scan_history_view.scan_selected.connect(self.scan_history_metadata_viewer.update_view)
        self.scan_history_view.scan_selected.connect(
            self.scan_history_device_viewer.update_devices_from_scan_history
        )
        self.scan_history_view.no_scan_selected.connect(
            self.scan_history_metadata_viewer.clear_view
        )
        self.scan_history_view.no_scan_selected.connect(self.scan_history_device_viewer.clear_view)


if __name__ == "__main__":  # pragma: no cover
    # pylint: disable=import-outside-toplevel

    from qtpy.QtWidgets import QApplication

    from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton

    app = QApplication([])
    main_window = QtWidgets.QMainWindow()

    central_widget = QtWidgets.QWidget()
    button = DarkModeButton()
    layout = QtWidgets.QVBoxLayout(central_widget)
    main_window.setCentralWidget(central_widget)
    # Create a ScanHistoryBrowser instance
    browser = ScanHistoryBrowser()  # type: ignore
    layout.addWidget(button)
    layout.addWidget(browser)
    main_window.setWindowTitle("Scan History Browser")
    main_window.resize(800, 400)
    main_window.show()
    app.exec_()



"""Module providing fuzzy search utilities for the BEC widgets."""

from __future__ import annotations

from typing import Any

from thefuzz import fuzz

FUZZY_SEARCH_THRESHOLD = 80


def is_match(
    text: str, row_data: dict[str, Any], relevant_keys: list[str], enable_fuzzy: bool
) -> bool:
    """
    Check if the text matches any of the relevant keys in the row data.

    Args:
        text (str): The text to search for.
        row_data (dict[str, Any]): The row data to search in.
        relevant_keys (list[str]): The keys to consider for searching.
        enable_fuzzy (bool): Whether to use fuzzy matching.
    Returns:
        bool: True if a match is found, False otherwise.
    """
    for key in relevant_keys:
        data = str(row_data.get(key, "") or "")
        if enable_fuzzy:
            match_ratio = fuzz.partial_ratio(text.lower(), data.lower())
            if match_ratio >= FUZZY_SEARCH_THRESHOLD:
                return True
        else:
            if text.lower() in data.lower():
                return True
    return False



from __future__ import annotations

import sys
from typing import Literal

import pyqtgraph as pg
from qtpy.QtCore import Property, QEasingCurve, QObject, QPropertyAnimation
from qtpy.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QMainWindow,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)
from typeguard import typechecked

from bec_widgets.widgets.containers.layout_manager.layout_manager import LayoutManagerWidget


class DimensionAnimator(QObject):
    """
    Helper class to animate the size of a panel widget.
    """

    def __init__(self, panel_widget: QWidget, direction: str):
        super().__init__()
        self.panel_widget = panel_widget
        self.direction = direction
        self._size = 0

    @Property(int)
    def panel_width(self):
        """
        Returns the current width of the panel widget.
        """
        return self._size

    @panel_width.setter
    def panel_width(self, val: int):
        """
        Set the width of the panel widget.

        Args:
            val(int): The width to set.
        """
        self._size = val
        self.panel_widget.setFixedWidth(val)

    @Property(int)
    def panel_height(self):
        """
        Returns the current height of the panel widget.
        """
        return self._size

    @panel_height.setter
    def panel_height(self, val: int):
        """
        Set the height of the panel widget.

        Args:
            val(int): The height to set.
        """
        self._size = val
        self.panel_widget.setFixedHeight(val)


class CollapsiblePanelManager(QObject):
    """
    Manager class to handle collapsible panels from a main widget using LayoutManagerWidget.
    """

    def __init__(self, layout_manager: LayoutManagerWidget, reference_widget: QWidget, parent=None):
        super().__init__(parent)
        self.layout_manager = layout_manager
        self.reference_widget = reference_widget
        self.animations = {}
        self.panels = {}
        self.direction_settings = {
            "left": {"property": b"maximumWidth", "default_size": 200},
            "right": {"property": b"maximumWidth", "default_size": 200},
            "top": {"property": b"maximumHeight", "default_size": 150},
            "bottom": {"property": b"maximumHeight", "default_size": 150},
        }

    def add_panel(
        self,
        direction: Literal["left", "right", "top", "bottom"],
        panel_widget: QWidget,
        target_size: int | None = None,
        duration: int = 300,
    ):
        """
        Add a panel widget to the layout manager.

        Args:
            direction(Literal["left", "right", "top", "bottom"]): Direction of the panel.
            panel_widget(QWidget): The panel widget to add.
            target_size(int, optional): The target size of the panel. Defaults to None.
            duration(int): The duration of the animation in milliseconds. Defaults to 300.
        """
        if direction not in self.direction_settings:
            raise ValueError("Direction must be one of 'left', 'right', 'top', 'bottom'.")

        if target_size is None:
            target_size = self.direction_settings[direction]["default_size"]

        self.layout_manager.add_widget_relative(
            widget=panel_widget, reference_widget=self.reference_widget, position=direction
        )
        panel_widget.setVisible(False)

        # Set initial constraints as flexible
        if direction in ["left", "right"]:
            panel_widget.setMaximumWidth(0)
            panel_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        else:
            panel_widget.setMaximumHeight(0)
            panel_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.panels[direction] = {
            "widget": panel_widget,
            "direction": direction,
            "target_size": target_size,
            "duration": duration,
            "animator": None,
        }

    def toggle_panel(
        self,
        direction: Literal["left", "right", "top", "bottom"],
        target_size: int | None = None,
        duration: int | None = None,
        easing_curve: QEasingCurve = QEasingCurve.InOutQuad,
        ensure_max: bool = False,
        scale: float | None = None,
        animation: bool = True,
    ):
        """
        Toggle the specified panel.

        Parameters:
            direction (Literal["left", "right", "top", "bottom"]): Direction of the panel to toggle.
            target_size (int, optional): Override target size for this toggle.
            duration (int, optional): Override the animation duration.
            easing_curve (QEasingCurve): Animation easing curve.
            ensure_max (bool): If True, animate as a fixed-size panel.
            scale (float, optional): If provided, calculate target_size from main widget size.
            animation (bool): If False, no animation is performed; panel instantly toggles.
        """
        if direction not in self.panels:
            raise ValueError(f"No panel found in direction '{direction}'.")

        panel_info = self.panels[direction]
        panel_widget = panel_info["widget"]
        dir_settings = self.direction_settings[direction]

        # Determine final target size
        if scale is not None:
            main_rect = self.reference_widget.geometry()
            if direction in ["left", "right"]:
                computed_target = int(main_rect.width() * scale)
            else:
                computed_target = int(main_rect.height() * scale)
            final_target_size = computed_target
        else:
            if target_size is None:
                final_target_size = panel_info["target_size"]
            else:
                final_target_size = target_size

        if duration is None:
            duration = panel_info["duration"]

        expanding_property = dir_settings["property"]
        currently_visible = panel_widget.isVisible()

        if ensure_max:
            if panel_info["animator"] is None:
                panel_info["animator"] = DimensionAnimator(panel_widget, direction)
            animator = panel_info["animator"]

            if direction in ["left", "right"]:
                prop_name = b"panel_width"
            else:
                prop_name = b"panel_height"
        else:
            animator = None
            prop_name = expanding_property

        if currently_visible:
            # Hide the panel
            if ensure_max:
                start_value = final_target_size
                end_value = 0
                finish_callback = lambda w=panel_widget, d=direction: self._after_hide_reset(w, d)
            else:
                start_value = (
                    panel_widget.width()
                    if direction in ["left", "right"]
                    else panel_widget.height()
                )
                end_value = 0
                finish_callback = lambda w=panel_widget: w.setVisible(False)
        else:
            # Show the panel
            start_value = 0
            end_value = final_target_size
            finish_callback = None
            if ensure_max:
                # Fix panel exactly
                if direction in ["left", "right"]:
                    panel_widget.setMinimumWidth(0)
                    panel_widget.setMaximumWidth(final_target_size)
                    panel_widget.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
                else:
                    panel_widget.setMinimumHeight(0)
                    panel_widget.setMaximumHeight(final_target_size)
                    panel_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
            else:
                # Flexible mode
                if direction in ["left", "right"]:
                    panel_widget.setMinimumWidth(0)
                    panel_widget.setMaximumWidth(final_target_size)
                    panel_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
                else:
                    panel_widget.setMinimumHeight(0)
                    panel_widget.setMaximumHeight(final_target_size)
                    panel_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

            panel_widget.setVisible(True)

        if not animation:
            # No animation: instantly set final state
            if end_value == 0:
                # Hiding
                if ensure_max:
                    # Reset after hide
                    self._after_hide_reset(panel_widget, direction)
                else:
                    panel_widget.setVisible(False)
            else:
                # Showing
                if ensure_max:
                    # Already set fixed size
                    if direction in ["left", "right"]:
                        panel_widget.setFixedWidth(end_value)
                    else:
                        panel_widget.setFixedHeight(end_value)
                else:
                    # Just set maximum dimension
                    if direction in ["left", "right"]:
                        panel_widget.setMaximumWidth(end_value)
                    else:
                        panel_widget.setMaximumHeight(end_value)
            return

        # With animation
        animation = QPropertyAnimation(animator if ensure_max else panel_widget, prop_name)
        animation.setDuration(duration)
        animation.setStartValue(start_value)
        animation.setEndValue(end_value)
        animation.setEasingCurve(easing_curve)

        if end_value == 0 and finish_callback:
            animation.finished.connect(finish_callback)
        elif end_value == 0 and not finish_callback:
            animation.finished.connect(lambda w=panel_widget: w.setVisible(False))

        animation.start()
        self.animations[panel_widget] = animation

    @typechecked
    def _after_hide_reset(
        self, panel_widget: QWidget, direction: Literal["left", "right", "top", "bottom"]
    ):
        """
        Reset the panel widget after hiding it in ensure_max mode.

        Args:
            panel_widget(QWidget): The panel widget to reset.
            direction(Literal["left", "right", "top", "bottom"]): The direction of the panel.
        """
        # Called after hiding a panel in ensure_max mode
        panel_widget.setVisible(False)
        if direction in ["left", "right"]:
            panel_widget.setMinimumWidth(0)
            panel_widget.setMaximumWidth(0)
            panel_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        else:
            panel_widget.setMinimumHeight(0)
            panel_widget.setMaximumHeight(16777215)
            panel_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)


####################################################################################################
# The following code is for the GUI control panel to interact with the CollapsiblePanelManager.
# It is not covered by any tests as it serves only as an example for the CollapsiblePanelManager class.
####################################################################################################


class MainWindow(QMainWindow):  # pragma: no cover
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Panels with ensure_max, scale, and animation toggle")
        self.resize(800, 600)

        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        main_layout = QVBoxLayout(central_widget)
        main_layout.setContentsMargins(10, 10, 10, 10)
        main_layout.setSpacing(10)

        # Buttons
        buttons_layout = QHBoxLayout()
        self.btn_left = QPushButton("Toggle Left (ensure_max=True)")
        self.btn_top = QPushButton("Toggle Top (scale=0.5, no animation)")
        self.btn_right = QPushButton("Toggle Right (ensure_max=True, scale=0.3)")
        self.btn_bottom = QPushButton("Toggle Bottom (no animation)")

        buttons_layout.addWidget(self.btn_left)
        buttons_layout.addWidget(self.btn_top)
        buttons_layout.addWidget(self.btn_right)
        buttons_layout.addWidget(self.btn_bottom)

        main_layout.addLayout(buttons_layout)

        self.layout_manager = LayoutManagerWidget()
        main_layout.addWidget(self.layout_manager)

        # Main widget
        self.main_plot = pg.PlotWidget()
        self.main_plot.plot([1, 2, 3, 4], [4, 3, 2, 1])
        self.layout_manager.add_widget(self.main_plot, 0, 0)

        self.panel_manager = CollapsiblePanelManager(self.layout_manager, self.main_plot)

        # Panels
        self.left_panel = pg.PlotWidget()
        self.left_panel.plot([1, 2, 3], [3, 2, 1])
        self.panel_manager.add_panel("left", self.left_panel, target_size=200)

        self.right_panel = pg.PlotWidget()
        self.right_panel.plot([10, 20, 30], [1, 10, 1])
        self.panel_manager.add_panel("right", self.right_panel, target_size=200)

        self.top_panel = pg.PlotWidget()
        self.top_panel.plot([1, 2, 3], [1, 2, 3])
        self.panel_manager.add_panel("top", self.top_panel, target_size=150)

        self.bottom_panel = pg.PlotWidget()
        self.bottom_panel.plot([2, 4, 6], [10, 5, 10])
        self.panel_manager.add_panel("bottom", self.bottom_panel, target_size=150)

        # Connect buttons
        # Left with ensure_max
        self.btn_left.clicked.connect(
            lambda: self.panel_manager.toggle_panel("left", ensure_max=True)
        )
        # Top with scale=0.5 and no animation
        self.btn_top.clicked.connect(
            lambda: self.panel_manager.toggle_panel("top", scale=0.5, animation=False)
        )
        # Right with ensure_max, scale=0.3
        self.btn_right.clicked.connect(
            lambda: self.panel_manager.toggle_panel("right", ensure_max=True, scale=0.3)
        )
        # Bottom no animation
        self.btn_bottom.clicked.connect(
            lambda: self.panel_manager.toggle_panel("bottom", target_size=100, animation=False)
        )


if __name__ == "__main__":  # pragma: no cover
    app = QApplication(sys.argv)
    w = MainWindow()
    w.show()
    sys.exit(app.exec())
